# The file was automatically generated by Lark v0.12.0
__version__ = "0.12.0"

#
#
#   Lark Stand-alone Generator Tool
# ----------------------------------
# Generates a stand-alone LALR(1) parser with a standard lexer
#
# Git:    https://github.com/erezsh/lark
# Author: Erez Shinan (erezshin@gmail.com)
#
#
#    >>> LICENSE
#
#    This tool and its generated code use a separate license from Lark,
#    and are subject to the terms of the Mozilla Public License, v. 2.0.
#    If a copy of the MPL was not distributed with this
#    file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
#    If you wish to purchase a commercial license for this tool and its
#    generated code, you may contact me via email or otherwise.
#
#    If MPL2 is incompatible with your free or open-source project,
#    contact me and we'll work it out.
#
#

from io import open



class LarkError(Exception):
    pass


class ConfigurationError(LarkError, ValueError):
    pass


def assert_config(value, options, msg='Got %r, expected one of %s'):
    if value not in options:
        raise ConfigurationError(msg % (value, options))


class GrammarError(LarkError):
    pass


class ParseError(LarkError):
    pass


class LexError(LarkError):
    pass


class UnexpectedInput(LarkError):
    #--
    pos_in_stream = None
    _terminals_by_name = None

    def get_context(self, text, span=40):
        #--
        assert self.pos_in_stream is not None, self
        pos = self.pos_in_stream
        start = max(pos - span, 0)
        end = pos + span
        if not isinstance(text, bytes):
            before = text[start:pos].rsplit('\n', 1)[-1]
            after = text[pos:end].split('\n', 1)[0]
            return before + after + '\n' + ' ' * len(before.expandtabs()) + '^\n'
        else:
            before = text[start:pos].rsplit(b'\n', 1)[-1]
            after = text[pos:end].split(b'\n', 1)[0]
            return (before + after + b'\n' + b' ' * len(before.expandtabs()) + b'^\n').decode("ascii", "backslashreplace")

    def match_examples(self, parse_fn, examples, token_type_match_fallback=False, use_accepts=False):
        #--
        assert self.state is not None, "Not supported for this exception"

        if isinstance(examples, dict):
            examples = examples.items()

        candidate = (None, False)
        for i, (label, example) in enumerate(examples):
            assert not isinstance(example, STRING_TYPE)

            for j, malformed in enumerate(example):
                try:
                    parse_fn(malformed)
                except UnexpectedInput as ut:
                    if ut.state == self.state:
                        if use_accepts and hasattr(self, 'accepts') and ut.accepts != self.accepts:
                            logger.debug("Different accepts with same state[%d]: %s != %s at example [%s][%s]" %
                                         (self.state, self.accepts, ut.accepts, i, j))
                            continue
                        try:
                            if ut.token == self.token:  ##

                                logger.debug("Exact Match at example [%s][%s]" % (i, j))
                                return label

                            if token_type_match_fallback:
                                ##

                                if (ut.token.type == self.token.type) and not candidate[-1]:
                                    logger.debug("Token Type Fallback at example [%s][%s]" % (i, j))
                                    candidate = label, True

                        except AttributeError:
                            pass
                        if candidate[0] is None:
                            logger.debug("Same State match at example [%s][%s]" % (i, j))
                            candidate = label, False

        return candidate[0]

    def _format_expected(self, expected):
        if self._terminals_by_name:
            d = self._terminals_by_name
            expected = [d[t_name].user_repr() if t_name in d else t_name for t_name in expected]
        return "Expected one of: \n\t* %s\n" % '\n\t* '.join(expected)


class UnexpectedEOF(ParseError, UnexpectedInput):
    #--

    def __init__(self, expected, state=None, terminals_by_name=None):
        super(UnexpectedEOF, self).__init__()

        self.expected = expected
        self.state = state
        from .lexer import Token
        self.token = Token("<EOF>", "")  ##

        self.pos_in_stream = -1
        self.line = -1
        self.column = -1
        self._terminals_by_name = terminals_by_name


    def __str__(self):
        message = "Unexpected end-of-input. "
        message += self._format_expected(self.expected)
        return message


class UnexpectedCharacters(LexError, UnexpectedInput):
    #--

    def __init__(self, seq, lex_pos, line, column, allowed=None, considered_tokens=None, state=None, token_history=None,
                 terminals_by_name=None, considered_rules=None):
        super(UnexpectedCharacters, self).__init__()

        ##

        self.line = line
        self.column = column
        self.pos_in_stream = lex_pos
        self.state = state
        self._terminals_by_name = terminals_by_name

        self.allowed = allowed
        self.considered_tokens = considered_tokens
        self.considered_rules = considered_rules
        self.token_history = token_history

        if isinstance(seq, bytes):
            self.char = seq[lex_pos:lex_pos + 1].decode("ascii", "backslashreplace")
        else:
            self.char = seq[lex_pos]
        self._context = self.get_context(seq)


    def __str__(self):
        message = "No terminal matches '%s' in the current parser context, at line %d col %d" % (self.char, self.line, self.column)
        message += '\n\n' + self._context
        if self.allowed:
            message += self._format_expected(self.allowed)
        if self.token_history:
            message += '\nPrevious tokens: %s\n' % ', '.join(repr(t) for t in self.token_history)
        return message


class UnexpectedToken(ParseError, UnexpectedInput):
    #--

    def __init__(self, token, expected, considered_rules=None, state=None, interactive_parser=None, terminals_by_name=None, token_history=None):
        super(UnexpectedToken, self).__init__()
        
        ##

        self.line = getattr(token, 'line', '?')
        self.column = getattr(token, 'column', '?')
        self.pos_in_stream = getattr(token, 'start_pos', None)
        self.state = state

        self.token = token
        self.expected = expected  ##

        self._accepts = NO_VALUE
        self.considered_rules = considered_rules
        self.interactive_parser = interactive_parser
        self._terminals_by_name = terminals_by_name
        self.token_history = token_history


    @property
    def accepts(self):
        if self._accepts is NO_VALUE:
            self._accepts = self.interactive_parser and self.interactive_parser.accepts()
        return self._accepts

    def __str__(self):
        message = ("Unexpected token %r at line %s, column %s.\n%s"
                   % (self.token, self.line, self.column, self._format_expected(self.accepts or self.expected)))
        if self.token_history:
            message += "Previous tokens: %r\n" % self.token_history

        return message

    @property
    def puppet(self):
        warn("UnexpectedToken.puppet attribute has been renamed to interactive_parser", DeprecationWarning)
        return self.interactive_parser
    


class VisitError(LarkError):
    #--

    def __init__(self, rule, obj, orig_exc):
        message = 'Error trying to process rule "%s":\n\n%s' % (rule, orig_exc)
        super(VisitError, self).__init__(message)

        self.rule = rule
        self.obj = obj
        self.orig_exc = orig_exc


import sys, re
import logging
from io import open
logger = logging.getLogger("lark")
logger.addHandler(logging.StreamHandler())
##

##

logger.setLevel(logging.CRITICAL)

if sys.version_info[0]>2:
    from abc import ABC, abstractmethod
else:
    from abc import ABCMeta, abstractmethod
    class ABC(object): ##

        __slots__ = ()
        __metclass__ = ABCMeta


Py36 = (sys.version_info[:2] >= (3, 6))

NO_VALUE = object()


def classify(seq, key=None, value=None):
    d = {}
    for item in seq:
        k = key(item) if (key is not None) else item
        v = value(item) if (value is not None) else item
        if k in d:
            d[k].append(v)
        else:
            d[k] = [v]
    return d


def _deserialize(data, namespace, memo):
    if isinstance(data, dict):
        if '__type__' in data:  ##

            class_ = namespace[data['__type__']]
            return class_.deserialize(data, memo)
        elif '@' in data:
            return memo[data['@']]
        return {key:_deserialize(value, namespace, memo) for key, value in data.items()}
    elif isinstance(data, list):
        return [_deserialize(value, namespace, memo) for value in data]
    return data


class Serialize(object):
    #--

    def memo_serialize(self, types_to_memoize):
        memo = SerializeMemoizer(types_to_memoize)
        return self.serialize(memo), memo.serialize()

    def serialize(self, memo=None):
        if memo and memo.in_types(self):
            return {'@': memo.memoized.get(self)}

        fields = getattr(self, '__serialize_fields__')
        res = {f: _serialize(getattr(self, f), memo) for f in fields}
        res['__type__'] = type(self).__name__
        if hasattr(self, '_serialize'):
            self._serialize(res, memo)
        return res

    @classmethod
    def deserialize(cls, data, memo):
        namespace = getattr(cls, '__serialize_namespace__', [])
        namespace = {c.__name__:c for c in namespace}

        fields = getattr(cls, '__serialize_fields__')

        if '@' in data:
            return memo[data['@']]

        inst = cls.__new__(cls)
        for f in fields:
            try:
                setattr(inst, f, _deserialize(data[f], namespace, memo))
            except KeyError as e:
                raise KeyError("Cannot find key for class", cls, e)

        if hasattr(inst, '_deserialize'):
            inst._deserialize()

        return inst


class SerializeMemoizer(Serialize):
    #--

    __serialize_fields__ = 'memoized',

    def __init__(self, types_to_memoize):
        self.types_to_memoize = tuple(types_to_memoize)
        self.memoized = Enumerator()

    def in_types(self, value):
        return isinstance(value, self.types_to_memoize)

    def serialize(self):
        return _serialize(self.memoized.reversed(), None)

    @classmethod
    def deserialize(cls, data, namespace, memo):
        return _deserialize(data, namespace, memo)


try:
    STRING_TYPE = basestring
except NameError:   ##

    STRING_TYPE = str


import types
from functools import wraps, partial
from contextlib import contextmanager

Str = type(u'')
try:
    classtype = types.ClassType  ##

except AttributeError:
    classtype = type    ##



def smart_decorator(f, create_decorator):
    if isinstance(f, types.FunctionType):
        return wraps(f)(create_decorator(f, True))

    elif isinstance(f, (classtype, type, types.BuiltinFunctionType)):
        return wraps(f)(create_decorator(f, False))

    elif isinstance(f, types.MethodType):
        return wraps(f)(create_decorator(f.__func__, True))

    elif isinstance(f, partial):
        ##

        return wraps(f.func)(create_decorator(lambda *args, **kw: f(*args[1:], **kw), True))

    else:
        return create_decorator(f.__func__.__call__, True)


try:
    import regex
except ImportError:
    regex = None

import sre_parse
import sre_constants
categ_pattern = re.compile(r'\\p{[A-Za-z_]+}')

def get_regexp_width(expr):
    if regex:
        ##

        ##

        ##

        regexp_final = re.sub(categ_pattern, 'A', expr)
    else:
        if re.search(categ_pattern, expr):
            raise ImportError('`regex` module must be installed in order to use Unicode categories.', expr)
        regexp_final = expr
    try:
        return [int(x) for x in sre_parse.parse(regexp_final).getwidth()]
    except sre_constants.error:
        if not regex:
            raise ValueError(expr)
        else:
            ##

            ##

            c = regex.compile(regexp_final)
            if c.match('') is None:
                return 1, sre_constants.MAXREPEAT
            else:
                return 0, sre_constants.MAXREPEAT


from collections import OrderedDict


class Meta:
    def __init__(self):
        self.empty = True


class Tree(object):
    #--
    def __init__(self, data, children, meta=None):
        self.data = data
        self.children = children
        self._meta = meta

    @property
    def meta(self):
        if self._meta is None:
            self._meta = Meta()
        return self._meta

    def __repr__(self):
        return 'Tree(%r, %r)' % (self.data, self.children)

    def _pretty_label(self):
        return self.data

    def _pretty(self, level, indent_str):
        if len(self.children) == 1 and not isinstance(self.children[0], Tree):
            return [indent_str*level, self._pretty_label(), '\t', '%s' % (self.children[0],), '\n']

        l = [indent_str*level, self._pretty_label(), '\n']
        for n in self.children:
            if isinstance(n, Tree):
                l += n._pretty(level+1, indent_str)
            else:
                l += [indent_str*(level+1), '%s' % (n,), '\n']

        return l

    def pretty(self, indent_str='  '):
        #--
        return ''.join(self._pretty(0, indent_str))

    def __eq__(self, other):
        try:
            return self.data == other.data and self.children == other.children
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.data, tuple(self.children)))

    def iter_subtrees(self):
        #--
        queue = [self]
        subtrees = OrderedDict()
        for subtree in queue:
            subtrees[id(subtree)] = subtree
            queue += [c for c in reversed(subtree.children)
                      if isinstance(c, Tree) and id(c) not in subtrees]

        del queue
        return reversed(list(subtrees.values()))

    def find_pred(self, pred):
        #--
        return filter(pred, self.iter_subtrees())

    def find_data(self, data):
        #--
        return self.find_pred(lambda t: t.data == data)


from inspect import getmembers, getmro


class Discard(Exception):
    #--
    pass

##



class _Decoratable:
    #--

    @classmethod
    def _apply_decorator(cls, decorator, **kwargs):
        mro = getmro(cls)
        assert mro[0] is cls
        libmembers = {name for _cls in mro[1:] for name, _ in getmembers(_cls)}
        for name, value in getmembers(cls):

            ##

            if name.startswith('_') or (name in libmembers and name not in cls.__dict__):
                continue
            if not callable(value):
                continue

            ##

            if hasattr(cls.__dict__[name], 'vargs_applied') or hasattr(value, 'vargs_applied'):
                continue

            static = isinstance(cls.__dict__[name], (staticmethod, classmethod))
            setattr(cls, name, decorator(value, static=static, **kwargs))
        return cls

    def __class_getitem__(cls, _):
        return cls


class Transformer(_Decoratable):
    #--
    __visit_tokens__ = True   ##


    def __init__(self,  visit_tokens=True):
        self.__visit_tokens__ = visit_tokens

    def _call_userfunc(self, tree, new_children=None):
        ##

        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            try:
                wrapper = getattr(f, 'visit_wrapper', None)
                if wrapper is not None:
                    return f.visit_wrapper(f, tree.data, children, tree.meta)
                else:
                    return f(children)
            except (GrammarError, Discard):
                raise
            except Exception as e:
                raise VisitError(tree.data, tree, e)

    def _call_userfunc_token(self, token):
        try:
            f = getattr(self, token.type)
        except AttributeError:
            return self.__default_token__(token)
        else:
            try:
                return f(token)
            except (GrammarError, Discard):
                raise
            except Exception as e:
                raise VisitError(token.type, token, e)

    def _transform_children(self, children):
        for c in children:
            try:
                if isinstance(c, Tree):
                    yield self._transform_tree(c)
                elif self.__visit_tokens__ and isinstance(c, Token):
                    yield self._call_userfunc_token(c)
                else:
                    yield c
            except Discard:
                pass

    def _transform_tree(self, tree):
        children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree, children)

    def transform(self, tree):
        #--
        return self._transform_tree(tree)

    def __mul__(self, other):
        #--
        return TransformerChain(self, other)

    def __default__(self, data, children, meta):
        #--
        return Tree(data, children, meta)

    def __default_token__(self, token):
        #--
        return token


def merge_transformers(base_transformer=None, **transformers_to_merge):
    #--
    if base_transformer is None:
        base_transformer = Transformer()
    for prefix, transformer in transformers_to_merge.items():
        for method_name in dir(transformer):
            method = getattr(transformer, method_name)
            if not callable(method):
                continue
            if method_name.startswith("_") or method_name == "transform":
                continue
            prefixed_method = prefix + "__" + method_name
            if hasattr(base_transformer, prefixed_method):
                raise AttributeError("Cannot merge: method '%s' appears more than once" % prefixed_method)

            setattr(base_transformer, prefixed_method, method)

    return base_transformer


class InlineTransformer(Transformer):   ##

    def _call_userfunc(self, tree, new_children=None):
        ##

        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)


class TransformerChain(object):
    def __init__(self, *transformers):
        self.transformers = transformers

    def transform(self, tree):
        for t in self.transformers:
            tree = t.transform(tree)
        return tree

    def __mul__(self, other):
        return TransformerChain(*self.transformers + (other,))


class Transformer_InPlace(Transformer):
    #--
    def _transform_tree(self, tree):           ##

        return self._call_userfunc(tree)

    def transform(self, tree):
        for subtree in tree.iter_subtrees():
            subtree.children = list(self._transform_children(subtree.children))

        return self._transform_tree(tree)


class Transformer_NonRecursive(Transformer):
    #--

    def transform(self, tree):
        ##

        rev_postfix = []
        q = [tree]
        while q:
            t = q.pop()
            rev_postfix.append(t)
            if isinstance(t, Tree):
                q += t.children

        ##

        stack = []
        for x in reversed(rev_postfix):
            if isinstance(x, Tree):
                size = len(x.children)
                if size:
                    args = stack[-size:]
                    del stack[-size:]
                else:
                    args = []
                stack.append(self._call_userfunc(x, args))
            elif self.__visit_tokens__ and isinstance(x, Token):
                stack.append(self._call_userfunc_token(x))
            else:
                stack.append(x)

        t ,= stack  ##

        return t


class Transformer_InPlaceRecursive(Transformer):
    #--
    def _transform_tree(self, tree):
        tree.children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree)


##


class VisitorBase:
    def _call_userfunc(self, tree):
        return getattr(self, tree.data, self.__default__)(tree)

    def __default__(self, tree):
        #--
        return tree

    def __class_getitem__(cls, _):
        return cls


class Visitor(VisitorBase):
    #--

    def visit(self, tree):
        #--
        for subtree in tree.iter_subtrees():
            self._call_userfunc(subtree)
        return tree

    def visit_topdown(self,tree):
        #--
        for subtree in tree.iter_subtrees_topdown():
            self._call_userfunc(subtree)
        return tree


class Visitor_Recursive(VisitorBase):
    #--

    def visit(self, tree):
        #--
        for child in tree.children:
            if isinstance(child, Tree):
                self.visit(child)

        self._call_userfunc(tree)
        return tree

    def visit_topdown(self,tree):
        #--
        self._call_userfunc(tree)

        for child in tree.children:
            if isinstance(child, Tree):
                self.visit_topdown(child)

        return tree


def visit_children_decor(func):
    #--
    @wraps(func)
    def inner(cls, tree):
        values = cls.visit_children(tree)
        return func(cls, values)
    return inner


class Interpreter(_Decoratable):
    #--

    def visit(self, tree):
        f = getattr(self, tree.data)
        wrapper = getattr(f, 'visit_wrapper', None)
        if wrapper is not None:
            return f.visit_wrapper(f, tree.data, tree.children, tree.meta)
        else:
            return f(tree)

    def visit_children(self, tree):
        return [self.visit(child) if isinstance(child, Tree) else child
                for child in tree.children]

    def __getattr__(self, name):
        return self.__default__

    def __default__(self, tree):
        return self.visit_children(tree)


##


def _apply_decorator(obj, decorator, **kwargs):
    try:
        _apply = obj._apply_decorator
    except AttributeError:
        return decorator(obj, **kwargs)
    else:
        return _apply(decorator, **kwargs)


def _inline_args__func(func):
    @wraps(func)
    def create_decorator(_f, with_self):
        if with_self:
            def f(self, children):
                return _f(self, *children)
        else:
            def f(self, children):
                return _f(*children)
        return f

    return smart_decorator(func, create_decorator)


def inline_args(obj):   ##

    return _apply_decorator(obj, _inline_args__func)


def _visitor_args_func_dec(func, visit_wrapper=None, static=False):
    def create_decorator(_f, with_self):
        if with_self:
            def f(self, *args, **kwargs):
                return _f(self, *args, **kwargs)
        else:
            def f(self, *args, **kwargs):
                return _f(*args, **kwargs)
        return f

    if static:
        f = wraps(func)(create_decorator(func, False))
    else:
        f = smart_decorator(func, create_decorator)
    f.vargs_applied = True
    f.visit_wrapper = visit_wrapper
    return f


def _vargs_inline(f, _data, children, _meta):
    return f(*children)
def _vargs_meta_inline(f, _data, children, meta):
    return f(meta, *children)
def _vargs_meta(f, _data, children, meta):
    return f(children, meta)   ##

def _vargs_tree(f, data, children, meta):
    return f(Tree(data, children, meta))


def v_args(inline=False, meta=False, tree=False, wrapper=None):
    #--
    if tree and (meta or inline):
        raise ValueError("Visitor functions cannot combine 'tree' with 'meta' or 'inline'.")

    func = None
    if meta:
        if inline:
            func = _vargs_meta_inline
        else:
            func = _vargs_meta
    elif inline:
        func = _vargs_inline
    elif tree:
        func = _vargs_tree

    if wrapper is not None:
        if func is not None:
            raise ValueError("Cannot use 'wrapper' along with 'tree', 'meta' or 'inline'.")
        func = wrapper

    def _visitor_args_dec(obj):
        return _apply_decorator(obj, _visitor_args_func_dec, visit_wrapper=func)
    return _visitor_args_dec




class Symbol(Serialize):
    __slots__ = ('name',)

    is_term = NotImplemented

    def __init__(self, name):
        self.name = name

    def __eq__(self, other):
        assert isinstance(other, Symbol), other
        return self.is_term == other.is_term and self.name == other.name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.name)

    fullrepr = property(__repr__)


class Terminal(Symbol):
    __serialize_fields__ = 'name', 'filter_out'

    is_term = True

    def __init__(self, name, filter_out=False):
        self.name = name
        self.filter_out = filter_out

    @property
    def fullrepr(self):
        return '%s(%r, %r)' % (type(self).__name__, self.name, self.filter_out)


class NonTerminal(Symbol):
    __serialize_fields__ = 'name',

    is_term = False


class RuleOptions(Serialize):
    __serialize_fields__ = 'keep_all_tokens', 'expand1', 'priority', 'template_source', 'empty_indices'

    def __init__(self, keep_all_tokens=False, expand1=False, priority=None, template_source=None, empty_indices=()):
        self.keep_all_tokens = keep_all_tokens
        self.expand1 = expand1
        self.priority = priority
        self.template_source = template_source
        self.empty_indices = empty_indices

    def __repr__(self):
        return 'RuleOptions(%r, %r, %r, %r)' % (
            self.keep_all_tokens,
            self.expand1,
            self.priority,
            self.template_source
        )


class Rule(Serialize):
    #--
    __slots__ = ('origin', 'expansion', 'alias', 'options', 'order', '_hash')

    __serialize_fields__ = 'origin', 'expansion', 'order', 'alias', 'options'
    __serialize_namespace__ = Terminal, NonTerminal, RuleOptions

    def __init__(self, origin, expansion, order=0, alias=None, options=None):
        self.origin = origin
        self.expansion = expansion
        self.alias = alias
        self.order = order
        self.options = options or RuleOptions()
        self._hash = hash((self.origin, tuple(self.expansion)))

    def _deserialize(self):
        self._hash = hash((self.origin, tuple(self.expansion)))

    def __str__(self):
        return '<%s : %s>' % (self.origin.name, ' '.join(x.name for x in self.expansion))

    def __repr__(self):
        return 'Rule(%r, %r, %r, %r)' % (self.origin, self.expansion, self.alias, self.options)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, Rule):
            return False
        return self.origin == other.origin and self.expansion == other.expansion



from warnings import warn
from copy import copy


class Pattern(Serialize):
    raw = None
    type = None

    def __init__(self, value, flags=(), raw=None):
        self.value = value
        self.flags = frozenset(flags)
        self.raw = raw

    def __repr__(self):
        return repr(self.to_regexp())

    ##

    def __hash__(self):
        return hash((type(self), self.value, self.flags))

    def __eq__(self, other):
        return type(self) == type(other) and self.value == other.value and self.flags == other.flags

    def to_regexp(self):
        raise NotImplementedError()

    def min_width(self):
        raise NotImplementedError()

    def max_width(self):
        raise NotImplementedError()

    if Py36:
        ##

        def _get_flags(self, value):
            for f in self.flags:
                value = ('(?%s:%s)' % (f, value))
            return value

    else:
        def _get_flags(self, value):
            for f in self.flags:
                value = ('(?%s)' % f) + value
            return value



class PatternStr(Pattern):
    __serialize_fields__ = 'value', 'flags'

    type = "str"

    def to_regexp(self):
        return self._get_flags(re.escape(self.value))

    @property
    def min_width(self):
        return len(self.value)
    max_width = min_width


class PatternRE(Pattern):
    __serialize_fields__ = 'value', 'flags', '_width'

    type = "re"

    def to_regexp(self):
        return self._get_flags(self.value)

    _width = None
    def _get_width(self):
        if self._width is None:
            self._width = get_regexp_width(self.to_regexp())
        return self._width

    @property
    def min_width(self):
        return self._get_width()[0]

    @property
    def max_width(self):
        return self._get_width()[1]


class TerminalDef(Serialize):
    __serialize_fields__ = 'name', 'pattern', 'priority'
    __serialize_namespace__ = PatternStr, PatternRE

    def __init__(self, name, pattern, priority=1):
        assert isinstance(pattern, Pattern), pattern
        self.name = name
        self.pattern = pattern
        self.priority = priority

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.name, self.pattern)

    def user_repr(self):
        if self.name.startswith('__'): ##

            return self.pattern.raw or self.name
        else:
            return self.name


class Token(Str):
    #--
    __slots__ = ('type', 'start_pos', 'value', 'line', 'column', 'end_line', 'end_column', 'end_pos')

    def __new__(cls, type_, value, start_pos=None, line=None, column=None, end_line=None, end_column=None, end_pos=None, pos_in_stream=None):
        try:
            inst = super(Token, cls).__new__(cls, value)
        except UnicodeDecodeError:
            value = value.decode('latin1')
            inst = super(Token, cls).__new__(cls, value)

        inst.type = type_
        inst.start_pos = start_pos if start_pos is not None else pos_in_stream
        inst.value = value
        inst.line = line
        inst.column = column
        inst.end_line = end_line
        inst.end_column = end_column
        inst.end_pos = end_pos
        return inst

    @property
    def pos_in_stream(self):
        warn("Attribute Token.pos_in_stream was renamed to Token.start_pos", DeprecationWarning, 2)
        return self.start_pos

    def update(self, type_=None, value=None):
        return Token.new_borrow_pos(
            type_ if type_ is not None else self.type,
            value if value is not None else self.value,
            self
        )

    @classmethod
    def new_borrow_pos(cls, type_, value, borrow_t):
        return cls(type_, value, borrow_t.start_pos, borrow_t.line, borrow_t.column, borrow_t.end_line, borrow_t.end_column, borrow_t.end_pos)

    def __reduce__(self):
        return (self.__class__, (self.type, self.value, self.start_pos, self.line, self.column))

    def __repr__(self):
        return 'Token(%r, %r)' % (self.type, self.value)

    def __deepcopy__(self, memo):
        return Token(self.type, self.value, self.start_pos, self.line, self.column)

    def __eq__(self, other):
        if isinstance(other, Token) and self.type != other.type:
            return False

        return Str.__eq__(self, other)

    __hash__ = Str.__hash__


class LineCounter:
    __slots__ = 'char_pos', 'line', 'column', 'line_start_pos', 'newline_char'

    def __init__(self, newline_char):
        self.newline_char = newline_char
        self.char_pos = 0
        self.line = 1
        self.column = 1
        self.line_start_pos = 0

    def __eq__(self, other):
        if not isinstance(other, LineCounter):
            return NotImplemented

        return self.char_pos == other.char_pos and self.newline_char == other.newline_char

    def feed(self, token, test_newline=True):
        #--
        if test_newline:
            newlines = token.count(self.newline_char)
            if newlines:
                self.line += newlines
                self.line_start_pos = self.char_pos + token.rindex(self.newline_char) + 1

        self.char_pos += len(token)
        self.column = self.char_pos - self.line_start_pos + 1


class UnlessCallback:
    def __init__(self, scanner):
        self.scanner = scanner

    def __call__(self, t):
        res = self.scanner.match(t.value, 0)
        if res:
            _value, t.type = res
        return t


class CallChain:
    def __init__(self, callback1, callback2, cond):
        self.callback1 = callback1
        self.callback2 = callback2
        self.cond = cond

    def __call__(self, t):
        t2 = self.callback1(t)
        return self.callback2(t) if self.cond(t2) else t2


def _get_match(re_, regexp, s, flags):
    m = re_.match(regexp, s, flags)
    if m:
        return m.group(0)

def _create_unless(terminals, g_regex_flags, re_, use_bytes):
    tokens_by_type = classify(terminals, lambda t: type(t.pattern))
    assert len(tokens_by_type) <= 2, tokens_by_type.keys()
    embedded_strs = set()
    callback = {}
    for retok in tokens_by_type.get(PatternRE, []):
        unless = []
        for strtok in tokens_by_type.get(PatternStr, []):
            if strtok.priority > retok.priority:
                continue
            s = strtok.pattern.value
            if s == _get_match(re_, retok.pattern.to_regexp(), s, g_regex_flags):
                unless.append(strtok)
                if strtok.pattern.flags <= retok.pattern.flags:
                    embedded_strs.add(strtok)
        if unless:
            callback[retok.name] = UnlessCallback(Scanner(unless, g_regex_flags, re_, match_whole=True, use_bytes=use_bytes))

    new_terminals = [t for t in terminals if t not in embedded_strs]
    return new_terminals, callback



class Scanner:
    def __init__(self, terminals, g_regex_flags, re_, use_bytes, match_whole=False):
        self.terminals = terminals
        self.g_regex_flags = g_regex_flags
        self.re_ = re_
        self.use_bytes = use_bytes
        self.match_whole = match_whole

        self.allowed_types = {t.name for t in self.terminals}

        self._mres = self._build_mres(terminals, len(terminals))

    def _build_mres(self, terminals, max_size):
        ##

        ##

        ##

        postfix = '$' if self.match_whole else ''
        mres = []
        while terminals:
            pattern = u'|'.join(u'(?P<%s>%s)' % (t.name, t.pattern.to_regexp() + postfix) for t in terminals[:max_size])
            if self.use_bytes:
                pattern = pattern.encode('latin-1')
            try:
                mre = self.re_.compile(pattern, self.g_regex_flags)
            except AssertionError:  ##

                return self._build_mres(terminals, max_size//2)

            mres.append((mre, {i: n for n, i in mre.groupindex.items()}))
            terminals = terminals[max_size:]
        return mres

    def match(self, text, pos):
        for mre, type_from_index in self._mres:
            m = mre.match(text, pos)
            if m:
                return m.group(0), type_from_index[m.lastindex]


def _regexp_has_newline(r):
    #--
    return '\n' in r or '\\n' in r or '\\s' in r or '[^' in r or ('(?s' in r and '.' in r)


class Lexer(object):
    #--
    lex = NotImplemented

    def make_lexer_state(self, text):
        line_ctr = LineCounter(b'\n' if isinstance(text, bytes) else '\n')
        return LexerState(text, line_ctr)


class TraditionalLexer(Lexer):

    def __init__(self, conf):
        terminals = list(conf.terminals)
        assert all(isinstance(t, TerminalDef) for t in terminals), terminals

        self.re = conf.re_module

        if not conf.skip_validation:
            ##

            for t in terminals:
                try:
                    self.re.compile(t.pattern.to_regexp(), conf.g_regex_flags)
                except self.re.error:
                    raise LexError("Cannot compile token %s: %s" % (t.name, t.pattern))

                if t.pattern.min_width == 0:
                    raise LexError("Lexer does not allow zero-width terminals. (%s: %s)" % (t.name, t.pattern))

            if not (set(conf.ignore) <= {t.name for t in terminals}):
                raise LexError("Ignore terminals are not defined: %s" % (set(conf.ignore) - {t.name for t in terminals}))

        ##

        self.newline_types = frozenset(t.name for t in terminals if _regexp_has_newline(t.pattern.to_regexp()))
        self.ignore_types = frozenset(conf.ignore)

        terminals.sort(key=lambda x: (-x.priority, -x.pattern.max_width, -len(x.pattern.value), x.name))
        self.terminals = terminals
        self.user_callbacks = conf.callbacks
        self.g_regex_flags = conf.g_regex_flags
        self.use_bytes = conf.use_bytes
        self.terminals_by_name = conf.terminals_by_name

        self._scanner = None

    def _build_scanner(self):
        terminals, self.callback = _create_unless(self.terminals, self.g_regex_flags, self.re, self.use_bytes)
        assert all(self.callback.values())

        for type_, f in self.user_callbacks.items():
            if type_ in self.callback:
                ##

                self.callback[type_] = CallChain(self.callback[type_], f, lambda t: t.type == type_)
            else:
                self.callback[type_] = f

        self._scanner = Scanner(terminals, self.g_regex_flags, self.re, self.use_bytes)

    @property
    def scanner(self):
        if self._scanner is None:
            self._build_scanner()
        return self._scanner

    def match(self, text, pos):
        return self.scanner.match(text, pos)

    def lex(self, state, parser_state):
        with suppress(EOFError):
            while True:
                yield self.next_token(state, parser_state)

    def next_token(self, lex_state, parser_state=None):
        line_ctr = lex_state.line_ctr
        while line_ctr.char_pos < len(lex_state.text):
            res = self.match(lex_state.text, line_ctr.char_pos)
            if not res:
                allowed = self.scanner.allowed_types - self.ignore_types
                if not allowed:
                    allowed = {"<END-OF-FILE>"}
                raise UnexpectedCharacters(lex_state.text, line_ctr.char_pos, line_ctr.line, line_ctr.column,
                                           allowed=allowed, token_history=lex_state.last_token and [lex_state.last_token],
                                           state=parser_state, terminals_by_name=self.terminals_by_name)

            value, type_ = res

            if type_ not in self.ignore_types:
                t = Token(type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column)
                line_ctr.feed(value, type_ in self.newline_types)
                t.end_line = line_ctr.line
                t.end_column = line_ctr.column
                t.end_pos = line_ctr.char_pos
                if t.type in self.callback:
                    t = self.callback[t.type](t)
                    if not isinstance(t, Token):
                        raise LexError("Callbacks must return a token (returned %r)" % t)
                lex_state.last_token = t
                return t
            else:
                if type_ in self.callback:
                    t2 = Token(type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column)
                    self.callback[type_](t2)
                line_ctr.feed(value, type_ in self.newline_types)

        ##

        raise EOFError(self)


class LexerState(object):
    __slots__ = 'text', 'line_ctr', 'last_token'

    def __init__(self, text, line_ctr, last_token=None):
        self.text = text
        self.line_ctr = line_ctr
        self.last_token = last_token

    def __eq__(self, other):
        if not isinstance(other, LexerState):
            return NotImplemented

        return self.text is other.text and self.line_ctr == other.line_ctr and self.last_token == other.last_token

    def __copy__(self):
        return type(self)(self.text, copy(self.line_ctr), self.last_token)


class ContextualLexer(Lexer):

    def __init__(self, conf, states, always_accept=()):
        terminals = list(conf.terminals)
        terminals_by_name = conf.terminals_by_name

        trad_conf = copy(conf)
        trad_conf.terminals = terminals

        lexer_by_tokens = {}
        self.lexers = {}
        for state, accepts in states.items():
            key = frozenset(accepts)
            try:
                lexer = lexer_by_tokens[key]
            except KeyError:
                accepts = set(accepts) | set(conf.ignore) | set(always_accept)
                lexer_conf = copy(trad_conf)
                lexer_conf.terminals = [terminals_by_name[n] for n in accepts if n in terminals_by_name]
                lexer = TraditionalLexer(lexer_conf)
                lexer_by_tokens[key] = lexer

            self.lexers[state] = lexer

        assert trad_conf.terminals is terminals
        self.root_lexer = TraditionalLexer(trad_conf)

    def make_lexer_state(self, text):
        return self.root_lexer.make_lexer_state(text)

    def lex(self, lexer_state, parser_state):
        try:
            while True:
                lexer = self.lexers[parser_state.position]
                yield lexer.next_token(lexer_state, parser_state)
        except EOFError:
            pass
        except UnexpectedCharacters as e:
            ##

            ##

            try:
                last_token = lexer_state.last_token  ##

                token = self.root_lexer.next_token(lexer_state, parser_state)
                raise UnexpectedToken(token, e.allowed, state=parser_state, token_history=[last_token], terminals_by_name=self.root_lexer.terminals_by_name)
            except UnexpectedCharacters:
                raise e  ##


class LexerThread(object):
    #--

    def __init__(self, lexer, text):
        self.lexer = lexer
        self.state = lexer.make_lexer_state(text)

    def lex(self, parser_state):
        return self.lexer.lex(self.state, parser_state)

    def __copy__(self):
        copied = object.__new__(LexerThread)
        copied.lexer = self.lexer
        copied.state = copy(self.state)
        return copied



class LexerConf(Serialize):
    __serialize_fields__ = 'terminals', 'ignore', 'g_regex_flags', 'use_bytes', 'lexer_type'
    __serialize_namespace__ = TerminalDef,

    def __init__(self, terminals, re_module, ignore=(), postlex=None, callbacks=None, g_regex_flags=0, skip_validation=False, use_bytes=False):
        self.terminals = terminals
        self.terminals_by_name = {t.name: t for t in self.terminals}
        assert len(self.terminals) == len(self.terminals_by_name)
        self.ignore = ignore
        self.postlex = postlex
        self.callbacks = callbacks or {}
        self.g_regex_flags = g_regex_flags
        self.re_module = re_module
        self.skip_validation = skip_validation
        self.use_bytes = use_bytes
        self.lexer_type = None

    @property
    def tokens(self):
        warn("LexerConf.tokens is deprecated. Use LexerConf.terminals instead", DeprecationWarning)
        return self.terminals

    def _deserialize(self):
        self.terminals_by_name = {t.name: t for t in self.terminals}

    def __deepcopy__(self, memo=None):
        return type(self)(
            deepcopy(self.terminals, memo),
            self.re_module,
            deepcopy(self.ignore, memo),
            deepcopy(self.postlex, memo),
            deepcopy(self.callbacks, memo),
            deepcopy(self.g_regex_flags, memo),
            deepcopy(self.skip_validation, memo),
            deepcopy(self.use_bytes, memo),
        )


class ParserConf(Serialize):
    __serialize_fields__ = 'rules', 'start', 'parser_type'

    def __init__(self, rules, callbacks, start):
        assert isinstance(start, list)
        self.rules = rules
        self.callbacks = callbacks
        self.start = start

        self.parser_type = None


from functools import partial, wraps
from itertools import repeat, product


class ExpandSingleChild:
    def __init__(self, node_builder):
        self.node_builder = node_builder

    def __call__(self, children):
        if len(children) == 1:
            return children[0]
        else:
            return self.node_builder(children)



class PropagatePositions:
    def __init__(self, node_builder, node_filter=None):
        self.node_builder = node_builder
        self.node_filter = node_filter

    def __call__(self, children):
        res = self.node_builder(children)

        if isinstance(res, Tree):
            ##

            ##

            ##

            ##


            res_meta = res.meta

            first_meta = self._pp_get_meta(children)
            if first_meta is not None:
                if not hasattr(res_meta, 'line'):
                    ##

                    res_meta.line = getattr(first_meta, 'container_line', first_meta.line)
                    res_meta.column = getattr(first_meta, 'container_column', first_meta.column)
                    res_meta.start_pos = getattr(first_meta, 'container_start_pos', first_meta.start_pos)
                    res_meta.empty = False

                res_meta.container_line = getattr(first_meta, 'container_line', first_meta.line)
                res_meta.container_column = getattr(first_meta, 'container_column', first_meta.column)

            last_meta = self._pp_get_meta(reversed(children))
            if last_meta is not None:
                if not hasattr(res_meta, 'end_line'):
                    res_meta.end_line = getattr(last_meta, 'container_end_line', last_meta.end_line)
                    res_meta.end_column = getattr(last_meta, 'container_end_column', last_meta.end_column)
                    res_meta.end_pos = getattr(last_meta, 'container_end_pos', last_meta.end_pos)
                    res_meta.empty = False

                res_meta.container_end_line = getattr(last_meta, 'container_end_line', last_meta.end_line)
                res_meta.container_end_column = getattr(last_meta, 'container_end_column', last_meta.end_column)

        return res

    def _pp_get_meta(self, children):
        for c in children:
            if self.node_filter is not None and not self.node_filter(c):
                continue
            if isinstance(c, Tree):
                if not c.meta.empty:
                    return c.meta
            elif isinstance(c, Token):
                return c

def make_propagate_positions(option):
    if callable(option):
        return partial(PropagatePositions, node_filter=option)
    elif option is True:
        return PropagatePositions
    elif option is False:
        return None

    raise ConfigurationError('Invalid option for propagate_positions: %r' % option)


class ChildFilter:
    def __init__(self, to_include, append_none, node_builder):
        self.node_builder = node_builder
        self.to_include = to_include
        self.append_none = append_none

    def __call__(self, children):
        filtered = []

        for i, to_expand, add_none in self.to_include:
            if add_none:
                filtered += [None] * add_none
            if to_expand:
                filtered += children[i].children
            else:
                filtered.append(children[i])

        if self.append_none:
            filtered += [None] * self.append_none

        return self.node_builder(filtered)


class ChildFilterLALR(ChildFilter):
    #--

    def __call__(self, children):
        filtered = []
        for i, to_expand, add_none in self.to_include:
            if add_none:
                filtered += [None] * add_none
            if to_expand:
                if filtered:
                    filtered += children[i].children
                else:   ##

                    filtered = children[i].children
            else:
                filtered.append(children[i])

        if self.append_none:
            filtered += [None] * self.append_none

        return self.node_builder(filtered)


class ChildFilterLALR_NoPlaceholders(ChildFilter):
    #--
    def __init__(self, to_include, node_builder):
        self.node_builder = node_builder
        self.to_include = to_include

    def __call__(self, children):
        filtered = []
        for i, to_expand in self.to_include:
            if to_expand:
                if filtered:
                    filtered += children[i].children
                else:   ##

                    filtered = children[i].children
            else:
                filtered.append(children[i])
        return self.node_builder(filtered)


def _should_expand(sym):
    return not sym.is_term and sym.name.startswith('_')


def maybe_create_child_filter(expansion, keep_all_tokens, ambiguous, _empty_indices):
    ##

    if _empty_indices:
        assert _empty_indices.count(False) == len(expansion)
        s = ''.join(str(int(b)) for b in _empty_indices)
        empty_indices = [len(ones) for ones in s.split('0')]
        assert len(empty_indices) == len(expansion)+1, (empty_indices, len(expansion))
    else:
        empty_indices = [0] * (len(expansion)+1)

    to_include = []
    nones_to_add = 0
    for i, sym in enumerate(expansion):
        nones_to_add += empty_indices[i]
        if keep_all_tokens or not (sym.is_term and sym.filter_out):
            to_include.append((i, _should_expand(sym), nones_to_add))
            nones_to_add = 0

    nones_to_add += empty_indices[len(expansion)]

    if _empty_indices or len(to_include) < len(expansion) or any(to_expand for i, to_expand,_ in to_include):
        if _empty_indices or ambiguous:
            return partial(ChildFilter if ambiguous else ChildFilterLALR, to_include, nones_to_add)
        else:
            ##

            return partial(ChildFilterLALR_NoPlaceholders, [(i, x) for i,x,_ in to_include])


class AmbiguousExpander:
    #--
    def __init__(self, to_expand, tree_class, node_builder):
        self.node_builder = node_builder
        self.tree_class = tree_class
        self.to_expand = to_expand

    def __call__(self, children):
        def _is_ambig_tree(t):
            return hasattr(t, 'data') and t.data == '_ambig'

        ##

        ##

        ##

        ##

        ambiguous = []
        for i, child in enumerate(children):
            if _is_ambig_tree(child):
                if i in self.to_expand:
                    ambiguous.append(i)

                child.expand_kids_by_data('_ambig')

        if not ambiguous:
            return self.node_builder(children)

        expand = [iter(child.children) if i in ambiguous else repeat(child) for i, child in enumerate(children)]
        return self.tree_class('_ambig', [self.node_builder(list(f[0])) for f in product(zip(*expand))])


def maybe_create_ambiguous_expander(tree_class, expansion, keep_all_tokens):
    to_expand = [i for i, sym in enumerate(expansion)
                 if keep_all_tokens or ((not (sym.is_term and sym.filter_out)) and _should_expand(sym))]
    if to_expand:
        return partial(AmbiguousExpander, to_expand, tree_class)


class AmbiguousIntermediateExpander:
    #--

    def __init__(self, tree_class, node_builder):
        self.node_builder = node_builder
        self.tree_class = tree_class

    def __call__(self, children):
        def _is_iambig_tree(child):
            return hasattr(child, 'data') and child.data == '_iambig'

        def _collapse_iambig(children):
            #--

            ##

            ##

            if children and _is_iambig_tree(children[0]):
                iambig_node = children[0]
                result = []
                for grandchild in iambig_node.children:
                    collapsed = _collapse_iambig(grandchild.children)
                    if collapsed:
                        for child in collapsed:
                            child.children += children[1:]
                        result += collapsed
                    else:
                        new_tree = self.tree_class('_inter', grandchild.children + children[1:])
                        result.append(new_tree)
                return result

        collapsed = _collapse_iambig(children)
        if collapsed:
            processed_nodes = [self.node_builder(c.children) for c in collapsed]
            return self.tree_class('_ambig', processed_nodes)

        return self.node_builder(children)


def ptb_inline_args(func):
    @wraps(func)
    def f(children):
        return func(*children)
    return f


def inplace_transformer(func):
    @wraps(func)
    def f(children):
        ##

        tree = Tree(func.__name__, children)
        return func(tree)
    return f


def apply_visit_wrapper(func, name, wrapper):
    if wrapper is _vargs_meta or wrapper is _vargs_meta_inline:
        raise NotImplementedError("Meta args not supported for internal transformer")

    @wraps(func)
    def f(children):
        return wrapper(func, name, children, None)
    return f


class ParseTreeBuilder:
    def __init__(self, rules, tree_class, propagate_positions=False, ambiguous=False, maybe_placeholders=False):
        self.tree_class = tree_class
        self.propagate_positions = propagate_positions
        self.ambiguous = ambiguous
        self.maybe_placeholders = maybe_placeholders

        self.rule_builders = list(self._init_builders(rules))

    def _init_builders(self, rules):
        propagate_positions = make_propagate_positions(self.propagate_positions)

        for rule in rules:
            options = rule.options
            keep_all_tokens = options.keep_all_tokens
            expand_single_child = options.expand1

            wrapper_chain = list(filter(None, [
                (expand_single_child and not rule.alias) and ExpandSingleChild,
                maybe_create_child_filter(rule.expansion, keep_all_tokens, self.ambiguous, options.empty_indices if self.maybe_placeholders else None),
                propagate_positions,
                self.ambiguous and maybe_create_ambiguous_expander(self.tree_class, rule.expansion, keep_all_tokens),
                self.ambiguous and partial(AmbiguousIntermediateExpander, self.tree_class)
            ]))

            yield rule, wrapper_chain

    def create_callback(self, transformer=None):
        callbacks = {}

        for rule, wrapper_chain in self.rule_builders:

            user_callback_name = rule.alias or rule.options.template_source or rule.origin.name
            try:
                f = getattr(transformer, user_callback_name)
                ##

                wrapper = getattr(f, 'visit_wrapper', None)
                if wrapper is not None:
                    f = apply_visit_wrapper(f, user_callback_name, wrapper)
                else:
                    if isinstance(transformer, InlineTransformer):
                        f = ptb_inline_args(f)
                    elif isinstance(transformer, Transformer_InPlace):
                        f = inplace_transformer(f)
            except AttributeError:
                f = partial(self.tree_class, user_callback_name)

            for w in wrapper_chain:
                f = w(f)

            if rule in callbacks:
                raise GrammarError("Rule '%s' already exists" % (rule,))

            callbacks[rule] = f

        return callbacks



class LALR_Parser(Serialize):
    def __init__(self, parser_conf, debug=False):
        analysis = LALR_Analyzer(parser_conf, debug=debug)
        analysis.compute_lalr()
        callbacks = parser_conf.callbacks

        self._parse_table = analysis.parse_table
        self.parser_conf = parser_conf
        self.parser = _Parser(analysis.parse_table, callbacks, debug)

    @classmethod
    def deserialize(cls, data, memo, callbacks, debug=False):
        inst = cls.__new__(cls)
        inst._parse_table = IntParseTable.deserialize(data, memo)
        inst.parser = _Parser(inst._parse_table, callbacks, debug)
        return inst

    def serialize(self, memo):
        return self._parse_table.serialize(memo)
    
    def parse_interactive(self, lexer, start):
        return self.parser.parse(lexer, start, start_interactive=True)

    def parse(self, lexer, start, on_error=None):
        try:
            return self.parser.parse(lexer, start)
        except UnexpectedInput as e:
            if on_error is None:
                raise

            while True:
                if isinstance(e, UnexpectedCharacters):
                    s = e.interactive_parser.lexer_state.state
                    p = s.line_ctr.char_pos

                if not on_error(e):
                    raise e

                if isinstance(e, UnexpectedCharacters):
                    ##

                    if p == s.line_ctr.char_pos:
                        s.line_ctr.feed(s.text[p:p+1])

                try:
                    return e.interactive_parser.resume_parse()
                except UnexpectedToken as e2:
                    if (isinstance(e, UnexpectedToken)
                        and e.token.type == e2.token.type == '$END'
                        and e.interactive_parser == e2.interactive_parser):
                        ##

                        raise e2
                    e = e2
                except UnexpectedCharacters as e2:
                    e = e2


class ParseConf(object):
    __slots__ = 'parse_table', 'callbacks', 'start', 'start_state', 'end_state', 'states'

    def __init__(self, parse_table, callbacks, start):
        self.parse_table = parse_table

        self.start_state = self.parse_table.start_states[start]
        self.end_state = self.parse_table.end_states[start]
        self.states = self.parse_table.states

        self.callbacks = callbacks
        self.start = start


class ParserState(object):
    __slots__ = 'parse_conf', 'lexer', 'state_stack', 'value_stack'

    def __init__(self, parse_conf, lexer, state_stack=None, value_stack=None):
        self.parse_conf = parse_conf
        self.lexer = lexer
        self.state_stack = state_stack or [self.parse_conf.start_state]
        self.value_stack = value_stack or []

    @property
    def position(self):
        return self.state_stack[-1]

    ##

    def __eq__(self, other):
        if not isinstance(other, ParserState):
            return NotImplemented
        return len(self.state_stack) == len(other.state_stack) and self.position == other.position

    def __copy__(self):
        return type(self)(
            self.parse_conf,
            self.lexer, ##

            copy(self.state_stack),
            deepcopy(self.value_stack),
        )

    def copy(self):
        return copy(self)

    def feed_token(self, token, is_end=False):
        state_stack = self.state_stack
        value_stack = self.value_stack
        states = self.parse_conf.states
        end_state = self.parse_conf.end_state
        callbacks = self.parse_conf.callbacks

        while True:
            state = state_stack[-1]
            try:
                action, arg = states[state][token.type]
            except KeyError:
                expected = {s for s in states[state].keys() if s.isupper()}
                raise UnexpectedToken(token, expected, state=self, interactive_parser=None)

            assert arg != end_state

            if action is Shift:
                ##

                assert not is_end
                state_stack.append(arg)
                value_stack.append(token if token.type not in callbacks else callbacks[token.type](token))
                return
            else:
                ##

                rule = arg
                size = len(rule.expansion)
                if size:
                    s = value_stack[-size:]
                    del state_stack[-size:]
                    del value_stack[-size:]
                else:
                    s = []

                value = callbacks[rule](s)

                _action, new_state = states[state_stack[-1]][rule.origin.name]
                assert _action is Shift
                state_stack.append(new_state)
                value_stack.append(value)

                if is_end and state_stack[-1] == end_state:
                    return value_stack[-1]

class _Parser(object):
    def __init__(self, parse_table, callbacks, debug=False):
        self.parse_table = parse_table
        self.callbacks = callbacks
        self.debug = debug

    def parse(self, lexer, start, value_stack=None, state_stack=None, start_interactive=False):
        parse_conf = ParseConf(self.parse_table, self.callbacks, start)
        parser_state = ParserState(parse_conf, lexer, state_stack, value_stack)
        if start_interactive:
            return InteractiveParser(self, parser_state, parser_state.lexer)
        return self.parse_from_state(parser_state)
    

    def parse_from_state(self, state):
        ##

        try:
            token = None
            for token in state.lexer.lex(state):
                state.feed_token(token)

            end_token = Token.new_borrow_pos('$END', '', token) if token else Token('$END', '', 0, 1, 1)
            return state.feed_token(end_token, True)
        except UnexpectedInput as e:
            try:
                e.interactive_parser = InteractiveParser(self, state, state.lexer)
            except NameError:
                pass
            raise e
        except Exception as e:
            if self.debug:
                print("")
                print("STATE STACK DUMP")
                print("----------------")
                for i, s in enumerate(state.state_stack):
                    print('%d)' % i , s)
                print("")

            raise


class Action:
    def __init__(self, name):
        self.name = name
    def __str__(self):
        return self.name
    def __repr__(self):
        return str(self)

Shift = Action('Shift')
Reduce = Action('Reduce')


class ParseTable:
    def __init__(self, states, start_states, end_states):
        self.states = states
        self.start_states = start_states
        self.end_states = end_states

    def serialize(self, memo):
        tokens = Enumerator()
        rules = Enumerator()

        states = {
            state: {tokens.get(token): ((1, arg.serialize(memo)) if action is Reduce else (0, arg))
                    for token, (action, arg) in actions.items()}
            for state, actions in self.states.items()
        }

        return {
            'tokens': tokens.reversed(),
            'states': states,
            'start_states': self.start_states,
            'end_states': self.end_states,
        }

    @classmethod
    def deserialize(cls, data, memo):
        tokens = data['tokens']
        states = {
            state: {tokens[token]: ((Reduce, Rule.deserialize(arg, memo)) if action==1 else (Shift, arg))
                    for token, (action, arg) in actions.items()}
            for state, actions in data['states'].items()
        }
        return cls(states, data['start_states'], data['end_states'])


class IntParseTable(ParseTable):

    @classmethod
    def from_ParseTable(cls, parse_table):
        enum = list(parse_table.states)
        state_to_idx = {s:i for i,s in enumerate(enum)}
        int_states = {}

        for s, la in parse_table.states.items():
            la = {k:(v[0], state_to_idx[v[1]]) if v[0] is Shift else v
                  for k,v in la.items()}
            int_states[ state_to_idx[s] ] = la


        start_states = {start:state_to_idx[s] for start, s in parse_table.start_states.items()}
        end_states = {start:state_to_idx[s] for start, s in parse_table.end_states.items()}
        return cls(int_states, start_states, end_states)



def _wrap_lexer(lexer_class):
    future_interface = getattr(lexer_class, '__future_interface__', False)
    if future_interface:
        return lexer_class
    else:
        class CustomLexerWrapper(Lexer):
            def __init__(self, lexer_conf):
                self.lexer = lexer_class(lexer_conf)
            def lex(self, lexer_state, parser_state):
                return self.lexer.lex(lexer_state.text)
        return CustomLexerWrapper


class MakeParsingFrontend:
    def __init__(self, parser_type, lexer_type):
        self.parser_type = parser_type
        self.lexer_type = lexer_type

    def __call__(self, lexer_conf, parser_conf, options):
        assert isinstance(lexer_conf, LexerConf)
        assert isinstance(parser_conf, ParserConf)
        parser_conf.parser_type = self.parser_type
        lexer_conf.lexer_type = self.lexer_type
        return ParsingFrontend(lexer_conf, parser_conf, options)

    def deserialize(self, data, memo, lexer_conf, callbacks, options):
        parser_conf = ParserConf.deserialize(data['parser_conf'], memo)
        parser = LALR_Parser.deserialize(data['parser'], memo, callbacks, options.debug)
        parser_conf.callbacks = callbacks
        return ParsingFrontend(lexer_conf, parser_conf, options, parser=parser)




class ParsingFrontend(Serialize):
    __serialize_fields__ = 'lexer_conf', 'parser_conf', 'parser', 'options'

    def __init__(self, lexer_conf, parser_conf, options, parser=None):
        self.parser_conf = parser_conf
        self.lexer_conf = lexer_conf
        self.options = options

        ##

        if parser:  ##

            self.parser = parser
        else:
            create_parser = {
                'lalr': create_lalr_parser,
                'earley': create_earley_parser,
                'cyk': CYK_FrontEnd,
            }[parser_conf.parser_type]
            self.parser = create_parser(lexer_conf, parser_conf, options)

        ##

        lexer_type = lexer_conf.lexer_type
        self.skip_lexer = False
        if lexer_type in ('dynamic', 'dynamic_complete'):
            assert lexer_conf.postlex is None
            self.skip_lexer = True
            return

        try:
            create_lexer = {
                'standard': create_traditional_lexer,
                'contextual': create_contextual_lexer,
            }[lexer_type]
        except KeyError:
            assert issubclass(lexer_type, Lexer), lexer_type
            self.lexer = _wrap_lexer(lexer_type)(lexer_conf)
        else:
            self.lexer = create_lexer(lexer_conf, self.parser, lexer_conf.postlex)

        if lexer_conf.postlex:
            self.lexer = PostLexConnector(self.lexer, lexer_conf.postlex)
    
    def _verify_start(self, start=None):
        if start is None:
            start_decls = self.parser_conf.start
            if len(start_decls) > 1:
                raise ConfigurationError("Lark initialized with more than 1 possible start rule. Must specify which start rule to parse", start_decls)
            start ,= start_decls
        elif start not in self.parser_conf.start:
            raise ConfigurationError("Unknown start rule %s. Must be one of %r" % (start, self.parser_conf.start))
        return start

    def parse(self, text, start=None, on_error=None):
        chosen_start = self._verify_start(start)
        stream = text if self.skip_lexer else LexerThread(self.lexer, text)
        kw = {} if on_error is None else {'on_error': on_error}
        return self.parser.parse(stream, chosen_start, **kw)
    
    def parse_interactive(self, text=None, start=None):
        chosen_start = self._verify_start(start)
        if self.parser_conf.parser_type != 'lalr':
            raise ConfigurationError("parse_interactive() currently only works with parser='lalr' ")
        stream = text if self.skip_lexer else LexerThread(self.lexer, text)
        return self.parser.parse_interactive(stream, chosen_start)


def get_frontend(parser, lexer):
    assert_config(parser, ('lalr', 'earley', 'cyk'))
    if not isinstance(lexer, type):     ##

        expected = {
            'lalr': ('standard', 'contextual'),
            'earley': ('standard', 'dynamic', 'dynamic_complete'),
            'cyk': ('standard', ),
         }[parser]
        assert_config(lexer, expected, 'Parser %r does not support lexer %%r, expected one of %%s' % parser)

    return MakeParsingFrontend(parser, lexer)


def _get_lexer_callbacks(transformer, terminals):
    result = {}
    for terminal in terminals:
        callback = getattr(transformer, terminal.name, None)
        if callback is not None:
            result[terminal.name] = callback
    return result

class PostLexConnector:
    def __init__(self, lexer, postlexer):
        self.lexer = lexer
        self.postlexer = postlexer

    def make_lexer_state(self, text):
        return self.lexer.make_lexer_state(text)

    def lex(self, lexer_state, parser_state):
        i = self.lexer.lex(lexer_state, parser_state)
        return self.postlexer.process(i)



def create_traditional_lexer(lexer_conf, parser, postlex):
    return TraditionalLexer(lexer_conf)

def create_contextual_lexer(lexer_conf, parser, postlex):
    states = {idx:list(t.keys()) for idx, t in parser._parse_table.states.items()}
    always_accept = postlex.always_accept if postlex else ()
    return ContextualLexer(lexer_conf, states, always_accept=always_accept)

def create_lalr_parser(lexer_conf, parser_conf, options=None):
    debug = options.debug if options else False
    return LALR_Parser(parser_conf, debug=debug)


create_earley_parser = NotImplemented
CYK_FrontEnd = NotImplemented



class LarkOptions(Serialize):
    #--
    OPTIONS_DOC = """
    **===  General Options  ===**

    start
            The start symbol. Either a string, or a list of strings for multiple possible starts (Default: "start")
    debug
            Display debug information and extra warnings. Use only when debugging (default: False)
            When used with Earley, it generates a forest graph as "sppf.png", if 'dot' is installed.
    transformer
            Applies the transformer to every parse tree (equivalent to applying it after the parse, but faster)
    propagate_positions
            Propagates (line, column, end_line, end_column) attributes into all tree branches.
            Accepts ``False``, ``True``, or a callable, which will filter which nodes to ignore when propagating.
    maybe_placeholders
            When ``True``, the ``[]`` operator returns ``None`` when not matched.

            When ``False``,  ``[]`` behaves like the ``?`` operator, and returns no value at all.
            (default= ``False``. Recommended to set to ``True``)
    cache
            Cache the results of the Lark grammar analysis, for x2 to x3 faster loading. LALR only for now.

            - When ``False``, does nothing (default)
            - When ``True``, caches to a temporary file in the local directory
            - When given a string, caches to the path pointed by the string
    regex
            When True, uses the ``regex`` module instead of the stdlib ``re``.
    g_regex_flags
            Flags that are applied to all terminals (both regex and strings)
    keep_all_tokens
            Prevent the tree builder from automagically removing "punctuation" tokens (default: False)
    tree_class
            Lark will produce trees comprised of instances of this class instead of the default ``lark.Tree``.

    **=== Algorithm Options ===**

    parser
            Decides which parser engine to use. Accepts "earley" or "lalr". (Default: "earley").
            (there is also a "cyk" option for legacy)
    lexer
            Decides whether or not to use a lexer stage

            - "auto" (default): Choose for me based on the parser
            - "standard": Use a standard lexer
            - "contextual": Stronger lexer (only works with parser="lalr")
            - "dynamic": Flexible and powerful (only with parser="earley")
            - "dynamic_complete": Same as dynamic, but tries *every* variation of tokenizing possible.
    ambiguity
            Decides how to handle ambiguity in the parse. Only relevant if parser="earley"

            - "resolve": The parser will automatically choose the simplest derivation
              (it chooses consistently: greedy for tokens, non-greedy for rules)
            - "explicit": The parser will return all derivations wrapped in "_ambig" tree nodes (i.e. a forest).
            - "forest": The parser will return the root of the shared packed parse forest.

    **=== Misc. / Domain Specific Options ===**

    postlex
            Lexer post-processing (Default: None) Only works with the standard and contextual lexers.
    priority
            How priorities should be evaluated - auto, none, normal, invert (Default: auto)
    lexer_callbacks
            Dictionary of callbacks for the lexer. May alter tokens during lexing. Use with caution.
    use_bytes
            Accept an input of type ``bytes`` instead of ``str`` (Python 3 only).
    edit_terminals
            A callback for editing the terminals before parse.
    import_paths
            A List of either paths or loader functions to specify from where grammars are imported
    source_path
            Override the source of from where the grammar was loaded. Useful for relative imports and unconventional grammar loading
    **=== End of Options ===**
    """
    if __doc__:
        __doc__ += OPTIONS_DOC


    ##

    ##

    ##

    ##

    ##

    ##

    ##

    ##

    _defaults = {
        'debug': False,
        'keep_all_tokens': False,
        'tree_class': None,
        'cache': False,
        'postlex': None,
        'parser': 'earley',
        'lexer': 'auto',
        'transformer': None,
        'start': 'start',
        'priority': 'auto',
        'ambiguity': 'auto',
        'regex': False,
        'propagate_positions': False,
        'lexer_callbacks': {},
        'maybe_placeholders': False,
        'edit_terminals': None,
        'g_regex_flags': 0,
        'use_bytes': False,
        'import_paths': [],
        'source_path': None,
    }

    def __init__(self, options_dict):
        o = dict(options_dict)

        options = {}
        for name, default in self._defaults.items():
            if name in o:
                value = o.pop(name)
                if isinstance(default, bool) and name not in ('cache', 'use_bytes', 'propagate_positions'):
                    value = bool(value)
            else:
                value = default

            options[name] = value

        if isinstance(options['start'], STRING_TYPE):
            options['start'] = [options['start']]

        self.__dict__['options'] = options


        assert_config(self.parser, ('earley', 'lalr', 'cyk', None))

        if self.parser == 'earley' and self.transformer:
            raise ConfigurationError('Cannot specify an embedded transformer when using the Earley algorithm. '
                             'Please use your transformer on the resulting parse tree, or use a different algorithm (i.e. LALR)')

        if o:
            raise ConfigurationError("Unknown options: %s" % o.keys())

    def __getattr__(self, name):
        try:
            return self.__dict__['options'][name]
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, name, value):
        assert_config(name, self.options.keys(), "%r isn't a valid option. Expected one of: %s")
        self.options[name] = value

    def serialize(self, memo):
        return self.options

    @classmethod
    def deserialize(cls, data, memo):
        return cls(data)


##

##

_LOAD_ALLOWED_OPTIONS = {'postlex', 'transformer', 'lexer_callbacks', 'use_bytes', 'debug', 'g_regex_flags', 'regex', 'propagate_positions', 'tree_class'}

_VALID_PRIORITY_OPTIONS = ('auto', 'normal', 'invert', None)
_VALID_AMBIGUITY_OPTIONS = ('auto', 'resolve', 'explicit', 'forest')


class PostLex(ABC):
    @abstractmethod
    def process(self, stream):
        return stream

    always_accept = ()


class Lark(Serialize):
    #--
    def __init__(self, grammar, **options):
        self.options = LarkOptions(options)

        ##

        use_regex = self.options.regex
        if use_regex:
            if regex:
                re_module = regex
            else:
                raise ImportError('`regex` module must be installed if calling `Lark(regex=True)`.')
        else:
            re_module = re

        ##

        if self.options.source_path is None:
            try:
                self.source_path = grammar.name
            except AttributeError:
                self.source_path = '<string>'
        else:
            self.source_path = self.options.source_path

        ##

        try:
            read = grammar.read
        except AttributeError:
            pass
        else:
            grammar = read()

        cache_fn = None
        cache_md5 = None
        if isinstance(grammar, STRING_TYPE):
            self.source_grammar = grammar
            if self.options.use_bytes:
                if not isascii(grammar):
                    raise ConfigurationError("Grammar must be ascii only, when use_bytes=True")
                if sys.version_info[0] == 2 and self.options.use_bytes != 'force':
                    raise ConfigurationError("`use_bytes=True` may have issues on python2."
                                              "Use `use_bytes='force'` to use it at your own risk.")

            if self.options.cache:
                if self.options.parser != 'lalr':
                    raise ConfigurationError("cache only works with parser='lalr' for now")

                unhashable = ('transformer', 'postlex', 'lexer_callbacks', 'edit_terminals')
                options_str = ''.join(k+str(v) for k, v in options.items() if k not in unhashable)
                from . import __version__
                s = grammar + options_str + __version__ + str(sys.version_info[:2])
                cache_md5 = hashlib.md5(s.encode('utf8')).hexdigest()

                if isinstance(self.options.cache, STRING_TYPE):
                    cache_fn = self.options.cache
                else:
                    if self.options.cache is not True:
                        raise ConfigurationError("cache argument must be bool or str")
                    ##

                    cache_fn = tempfile.gettempdir() + '/.lark_cache_%s_%s_%s.tmp' % ((cache_md5,) + sys.version_info[:2])

                if FS.exists(cache_fn):
                    logger.debug('Loading grammar from cache: %s', cache_fn)
                    ##

                    for name in (set(options) - _LOAD_ALLOWED_OPTIONS):
                        del options[name]
                    with FS.open(cache_fn, 'rb') as f:
                        old_options = self.options
                        try:
                            file_md5 = f.readline().rstrip(b'\n')
                            cached_used_files = pickle.load(f)
                            if file_md5 == cache_md5.encode('utf8') and verify_used_files(cached_used_files):
                                cached_parser_data = pickle.load(f)
                                self._load(cached_parser_data, **options)
                                return
                        except Exception: ##

                            logger.exception("Failed to load Lark from cache: %r. We will try to carry on." % cache_fn)
                            
                            ##

                            ##

                            self.options = old_options


            ##

            self.grammar, used_files = load_grammar(grammar, self.source_path, self.options.import_paths, self.options.keep_all_tokens)
        else:
            assert isinstance(grammar, Grammar)
            self.grammar = grammar


        if self.options.lexer == 'auto':
            if self.options.parser == 'lalr':
                self.options.lexer = 'contextual'
            elif self.options.parser == 'earley':
                if self.options.postlex is not None:
                    logger.info("postlex can't be used with the dynamic lexer, so we use standard instead. "
                                "Consider using lalr with contextual instead of earley")
                    self.options.lexer = 'standard'
                else:
                    self.options.lexer = 'dynamic'
            elif self.options.parser == 'cyk':
                self.options.lexer = 'standard'
            else:
                assert False, self.options.parser
        lexer = self.options.lexer
        if isinstance(lexer, type):
            assert issubclass(lexer, Lexer)     ##

        else:
            assert_config(lexer, ('standard', 'contextual', 'dynamic', 'dynamic_complete'))
            if self.options.postlex is not None and 'dynamic' in lexer:
                raise ConfigurationError("Can't use postlex with a dynamic lexer. Use standard or contextual instead")

        if self.options.ambiguity == 'auto':
            if self.options.parser == 'earley':
                self.options.ambiguity = 'resolve'
        else:
            assert_config(self.options.parser, ('earley', 'cyk'), "%r doesn't support disambiguation. Use one of these parsers instead: %s")

        if self.options.priority == 'auto':
            self.options.priority = 'normal'

        if self.options.priority not in _VALID_PRIORITY_OPTIONS:
            raise ConfigurationError("invalid priority option: %r. Must be one of %r" % (self.options.priority, _VALID_PRIORITY_OPTIONS))
        assert self.options.ambiguity not in ('resolve__antiscore_sum', ), 'resolve__antiscore_sum has been replaced with the option priority="invert"'
        if self.options.ambiguity not in _VALID_AMBIGUITY_OPTIONS:
            raise ConfigurationError("invalid ambiguity option: %r. Must be one of %r" % (self.options.ambiguity, _VALID_AMBIGUITY_OPTIONS))

        if self.options.parser is None:
            terminals_to_keep = '*'
        elif self.options.postlex is not None:
            terminals_to_keep = set(self.options.postlex.always_accept)
        else:
            terminals_to_keep = set()

        ##

        self.terminals, self.rules, self.ignore_tokens = self.grammar.compile(self.options.start, terminals_to_keep)

        if self.options.edit_terminals:
            for t in self.terminals:
                self.options.edit_terminals(t)

        self._terminals_dict = {t.name: t for t in self.terminals}

        ##

        ##

        if self.options.priority == 'invert':
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = -rule.options.priority
        ##

        ##

        ##

        elif self.options.priority is None:
            for rule in self.rules:
                if rule.options.priority is not None:
                    rule.options.priority = None

        ##

        self.lexer_conf = LexerConf(
                self.terminals, re_module, self.ignore_tokens, self.options.postlex,
                self.options.lexer_callbacks, self.options.g_regex_flags, use_bytes=self.options.use_bytes
            )

        if self.options.parser:
            self.parser = self._build_parser()
        elif lexer:
            self.lexer = self._build_lexer()

        if cache_fn:
            logger.debug('Saving grammar to cache: %s', cache_fn)
            with FS.open(cache_fn, 'wb') as f:
                f.write(cache_md5.encode('utf8') + b'\n')
                pickle.dump(used_files, f)
                self.save(f)

    if __doc__:
        __doc__ += "\n\n" + LarkOptions.OPTIONS_DOC

    __serialize_fields__ = 'parser', 'rules', 'options'

    def _build_lexer(self, dont_ignore=False):
        lexer_conf = self.lexer_conf
        if dont_ignore:
            from copy import copy
            lexer_conf = copy(lexer_conf)
            lexer_conf.ignore = ()
        return TraditionalLexer(lexer_conf)

    def _prepare_callbacks(self):
        self._callbacks = {}
        ##

        if self.options.ambiguity != 'forest':
            self._parse_tree_builder = ParseTreeBuilder(
                    self.rules,
                    self.options.tree_class or Tree,
                    self.options.propagate_positions,
                    self.options.parser != 'lalr' and self.options.ambiguity == 'explicit',
                    self.options.maybe_placeholders
                )
            self._callbacks = self._parse_tree_builder.create_callback(self.options.transformer)
        self._callbacks.update(_get_lexer_callbacks(self.options.transformer, self.terminals))

    def _build_parser(self):
        self._prepare_callbacks()
        parser_class = get_frontend(self.options.parser, self.options.lexer)
        parser_conf = ParserConf(self.rules, self._callbacks, self.options.start)
        return parser_class(self.lexer_conf, parser_conf, options=self.options)

    def save(self, f):
        #--
        data, m = self.memo_serialize([TerminalDef, Rule])
        pickle.dump({'data': data, 'memo': m}, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, f):
        #--
        inst = cls.__new__(cls)
        return inst._load(f)

    def _deserialize_lexer_conf(self, data, memo, options):
        lexer_conf = LexerConf.deserialize(data['lexer_conf'], memo)
        lexer_conf.callbacks = options.lexer_callbacks or {}
        lexer_conf.re_module = regex if options.regex else re
        lexer_conf.use_bytes = options.use_bytes
        lexer_conf.g_regex_flags = options.g_regex_flags
        lexer_conf.skip_validation = True
        lexer_conf.postlex = options.postlex
        return lexer_conf

    def _load(self, f, **kwargs):
        if isinstance(f, dict):
            d = f
        else:
            d = pickle.load(f)
        memo_json = d['memo']
        data = d['data']

        assert memo_json
        memo = SerializeMemoizer.deserialize(memo_json, {'Rule': Rule, 'TerminalDef': TerminalDef}, {})
        options = dict(data['options'])
        if (set(kwargs) - _LOAD_ALLOWED_OPTIONS) & set(LarkOptions._defaults):
            raise ConfigurationError("Some options are not allowed when loading a Parser: {}"
                             .format(set(kwargs) - _LOAD_ALLOWED_OPTIONS))
        options.update(kwargs)
        self.options = LarkOptions.deserialize(options, memo)
        self.rules = [Rule.deserialize(r, memo) for r in data['rules']]
        self.source_path = '<deserialized>'
        parser_class = get_frontend(self.options.parser, self.options.lexer)
        self.lexer_conf = self._deserialize_lexer_conf(data['parser'], memo, self.options)
        self.terminals = self.lexer_conf.terminals
        self._prepare_callbacks()
        self._terminals_dict = {t.name: t for t in self.terminals}
        self.parser = parser_class.deserialize(
            data['parser'],
            memo,
            self.lexer_conf,
            self._callbacks,
            self.options,  ##

        )
        return self

    @classmethod
    def _load_from_dict(cls, data, memo, **kwargs):
        inst = cls.__new__(cls)
        return inst._load({'data': data, 'memo': memo}, **kwargs)

    @classmethod
    def open(cls, grammar_filename, rel_to=None, **options):
        #--
        if rel_to:
            basepath = os.path.dirname(rel_to)
            grammar_filename = os.path.join(basepath, grammar_filename)
        with open(grammar_filename, encoding='utf8') as f:
            return cls(f, **options)

    @classmethod
    def open_from_package(cls, package, grammar_path, search_paths=("",), **options):
        #--
        package_loader = FromPackageLoader(package, search_paths)
        full_path, text = package_loader(None, grammar_path)
        options.setdefault('source_path', full_path)
        options.setdefault('import_paths', [])
        options['import_paths'].append(package_loader)
        return cls(text, **options)

    def __repr__(self):
        return 'Lark(open(%r), parser=%r, lexer=%r, ...)' % (self.source_path, self.options.parser, self.options.lexer)


    def lex(self, text, dont_ignore=False):
        #--
        if not hasattr(self, 'lexer') or dont_ignore:
            lexer = self._build_lexer(dont_ignore)
        else:
            lexer = self.lexer
        lexer_thread = LexerThread(lexer, text)
        stream = lexer_thread.lex(None)
        if self.options.postlex:
            return self.options.postlex.process(stream)
        return stream

    def get_terminal(self, name):
        #--
        return self._terminals_dict[name]
    
    def parse_interactive(self, text=None, start=None):
        #--
        return self.parser.parse_interactive(text, start=start)

    def parse(self, text, start=None, on_error=None):
        #--
        return self.parser.parse(text, start=start, on_error=on_error)

    @property
    def source(self):
        warn("Attribute Lark.source was renamed to Lark.source_path", DeprecationWarning)
        return self.source_path

    @source.setter
    def source(self, value):
        self.source_path = value

    @property
    def grammar_source(self):
        warn("Attribute Lark.grammar_source was renamed to Lark.source_grammar", DeprecationWarning)
        return self.source_grammar

    @grammar_source.setter
    def grammar_source(self, value):
        self.source_grammar = value



class DedentError(LarkError):
    pass

class Indenter(PostLex):
    def __init__(self):
        self.paren_level = None
        self.indent_level = None
        assert self.tab_len > 0

    def handle_NL(self, token):
        if self.paren_level > 0:
            return

        yield token

        indent_str = token.rsplit('\n', 1)[1] ##

        indent = indent_str.count(' ') + indent_str.count('\t') * self.tab_len

        if indent > self.indent_level[-1]:
            self.indent_level.append(indent)
            yield Token.new_borrow_pos(self.INDENT_type, indent_str, token)
        else:
            while indent < self.indent_level[-1]:
                self.indent_level.pop()
                yield Token.new_borrow_pos(self.DEDENT_type, indent_str, token)

            if indent != self.indent_level[-1]:
                raise DedentError('Unexpected dedent to column %s. Expected dedent to %s' % (indent, self.indent_level[-1]))

    def _process(self, stream):
        for token in stream:
            if token.type == self.NL_type:
                for t in self.handle_NL(token):
                    yield t
            else:
                yield token

            if token.type in self.OPEN_PAREN_types:
                self.paren_level += 1
            elif token.type in self.CLOSE_PAREN_types:
                self.paren_level -= 1
                assert self.paren_level >= 0

        while len(self.indent_level) > 1:
            self.indent_level.pop()
            yield Token(self.DEDENT_type, '')

        assert self.indent_level == [0], self.indent_level

    def process(self, stream):
        self.paren_level = 0
        self.indent_level = [0]
        return self._process(stream)

    ##

    @property
    def always_accept(self):
        return (self.NL_type,)


import pickle, zlib, base64
DATA = (
{'parser': {'lexer_conf': {'terminals': [{'@': 0}, {'@': 1}, {'@': 2}, {'@': 3}, {'@': 4}, {'@': 5}, {'@': 6}, {'@': 7}, {'@': 8}, {'@': 9}, {'@': 10}, {'@': 11}, {'@': 12}, {'@': 13}, {'@': 14}, {'@': 15}, {'@': 16}, {'@': 17}, {'@': 18}, {'@': 19}, {'@': 20}, {'@': 21}, {'@': 22}, {'@': 23}, {'@': 24}, {'@': 25}, {'@': 26}, {'@': 27}, {'@': 28}, {'@': 29}, {'@': 30}, {'@': 31}, {'@': 32}, {'@': 33}, {'@': 34}, {'@': 35}, {'@': 36}, {'@': 37}, {'@': 38}, {'@': 39}, {'@': 40}, {'@': 41}, {'@': 42}, {'@': 43}, {'@': 44}, {'@': 45}, {'@': 46}, {'@': 47}, {'@': 48}, {'@': 49}, {'@': 50}, {'@': 51}, {'@': 52}, {'@': 53}, {'@': 54}], 'ignore': ['WS'], 'g_regex_flags': 0, 'use_bytes': False, 'lexer_type': 'contextual', '__type__': 'LexerConf'}, 'parser_conf': {'rules': [{'@': 55}, {'@': 56}, {'@': 57}, {'@': 58}, {'@': 59}, {'@': 60}, {'@': 61}, {'@': 62}, {'@': 63}, {'@': 64}, {'@': 65}, {'@': 66}, {'@': 67}, {'@': 68}, {'@': 69}, {'@': 70}, {'@': 71}, {'@': 72}, {'@': 73}, {'@': 74}, {'@': 75}, {'@': 76}, {'@': 77}, {'@': 78}, {'@': 79}, {'@': 80}, {'@': 81}, {'@': 82}, {'@': 83}, {'@': 84}, {'@': 85}, {'@': 86}, {'@': 87}, {'@': 88}, {'@': 89}, {'@': 90}, {'@': 91}, {'@': 92}, {'@': 93}, {'@': 94}, {'@': 95}, {'@': 96}, {'@': 97}, {'@': 98}, {'@': 99}, {'@': 100}, {'@': 101}, {'@': 102}, {'@': 103}, {'@': 104}, {'@': 105}, {'@': 106}, {'@': 107}, {'@': 108}, {'@': 109}, {'@': 110}, {'@': 111}, {'@': 112}, {'@': 113}, {'@': 114}, {'@': 115}, {'@': 116}, {'@': 117}, {'@': 118}, {'@': 119}, {'@': 120}, {'@': 121}, {'@': 122}, {'@': 123}, {'@': 124}, {'@': 125}, {'@': 126}, {'@': 127}, {'@': 128}, {'@': 129}, {'@': 130}, {'@': 131}, {'@': 132}, {'@': 133}, {'@': 134}, {'@': 135}, {'@': 136}, {'@': 137}, {'@': 138}, {'@': 139}, {'@': 140}, {'@': 141}, {'@': 142}, {'@': 143}, {'@': 144}, {'@': 145}, {'@': 146}, {'@': 147}, {'@': 148}, {'@': 149}, {'@': 150}, {'@': 151}, {'@': 152}, {'@': 153}, {'@': 154}, {'@': 155}, {'@': 156}, {'@': 157}, {'@': 158}, {'@': 159}, {'@': 160}, {'@': 161}, {'@': 162}, {'@': 163}, {'@': 164}, {'@': 165}, {'@': 166}, {'@': 167}, {'@': 168}, {'@': 169}, {'@': 170}, {'@': 171}, {'@': 172}, {'@': 173}, {'@': 174}, {'@': 175}, {'@': 176}, {'@': 177}, {'@': 178}, {'@': 179}, {'@': 180}, {'@': 181}, {'@': 182}, {'@': 183}, {'@': 184}, {'@': 185}, {'@': 186}, {'@': 187}, {'@': 188}, {'@': 189}, {'@': 190}, {'@': 191}, {'@': 192}, {'@': 193}, {'@': 194}, {'@': 195}, {'@': 196}, {'@': 197}, {'@': 198}, {'@': 199}, {'@': 200}, {'@': 201}, {'@': 202}, {'@': 203}, {'@': 204}, {'@': 205}, {'@': 206}, {'@': 207}, {'@': 208}, {'@': 209}, {'@': 210}, {'@': 211}, {'@': 212}, {'@': 213}, {'@': 214}, {'@': 215}, {'@': 216}, {'@': 217}, {'@': 218}, {'@': 219}], 'start': ['start'], 'parser_type': 'lalr', '__type__': 'ParserConf'}, 'parser': {'tokens': {0: 'STAR', 1: 'DOT', 2: '__ANON_9', 3: '__ANON_11', 4: '__ANON_14', 5: 'BANG', 6: 'TILDE', 7: '__ANON_0', 8: '__ANON_13', 9: '__ANON_1', 10: '__ANON_6', 11: '__ANON_7', 12: 'COMMA', 13: 'AMPERSAND', 14: '__ANON_5', 15: 'QMARK', 16: '__ANON_3', 17: 'RBRACE', 18: '__ANON_15', 19: '__ANON_12', 20: '__ANON_2', 21: '__ANON_4', 22: '__ANON_8', 23: 'MORETHAN', 24: 'RPAR', 25: '__ANON_10', 26: 'MINUS', 27: 'RSQB', 28: 'AT', 29: 'VBAR', 30: '$END', 31: 'term_nonvar', 32: 'interval', 33: 'LPAR', 34: 'variable', 35: 'set', 36: 'term', 37: 'multi', 38: 'negation', 39: 'word', 40: 'compound_term', 41: 'HASH', 42: 'con_negation', 43: 'string', 44: 'con_int_set', 45: 'LESSTHAN', 46: 'statement', 47: 'PLUS', 48: 'string_raw', 49: '__ANON_16', 50: 'ext_image', 51: 'con_ext_set', 52: '__ANON_25', 53: 'single', 54: 'DOLLAR', 55: 'int_image', 56: 'CIRCUMFLEX', 57: '__ANON_24', 58: 'LSQB', 59: 'ext_set', 60: 'int_set', 61: 'LBRACE', 62: 'op', 63: '__statement_star_0', 64: 'term5', 65: 'multi_disj_expr', 66: 'multi_conj_expr', 67: 'term6', 68: 'term4', 69: 'multi_extint_expr', 70: 'multi_sequential_expr', 71: 'multi_infix_expr', 72: 'term3', 73: 'multi_prod_expr', 74: 'term2', 75: 'term1', 76: 'multi_parallel_expr', 77: 'multi_intint_expr', 78: 'SEMICOLON', 79: 'PERCENT', 80: '__ANON_18', 81: 'tense', 82: '__ANON_20', 83: 'truth', 84: '__ANON_19', 85: '__ANON_17', 86: 'k_evidence', 87: '__ANON_23', 88: 'desire', 89: '__multi_intint_expr_plus_6', 90: 'sentence', 91: 'task', 92: 'budget', 93: 'start', 94: 'priority', 95: '__ANON_21', 96: '__multi_sequential_expr_plus_4', 97: 'NUMBER', 98: 'confidence', 99: '__ANON_22', 100: 'con_single', 101: 'con_int_image', 102: 'con_ext_image', 103: 'con_multi', 104: 'SLASH', 105: 'con_product', 106: 'BACKSLASH', 107: 'copula', 108: '__multi_parallel_expr_plus_5', 109: 'quality', 110: 'durability', 111: '__multi_extint_expr_plus_7', 112: '__multi_disj_expr_plus_2', 113: '__multi_prod_expr_plus_1', 114: '__multi_conj_expr_plus_3', 115: 'frequency', 116: 'COLON'}, 'states': {0: {0: (1, {'@': 116}), 1: (1, {'@': 116}), 2: (1, {'@': 116}), 3: (1, {'@': 116}), 4: (1, {'@': 116}), 5: (1, {'@': 116}), 6: (1, {'@': 116}), 7: (1, {'@': 116}), 8: (1, {'@': 116}), 9: (1, {'@': 116}), 10: (1, {'@': 116}), 11: (1, {'@': 116}), 12: (1, {'@': 116}), 13: (1, {'@': 116}), 14: (1, {'@': 116}), 15: (1, {'@': 116}), 16: (1, {'@': 116}), 17: (1, {'@': 116}), 18: (1, {'@': 116}), 19: (1, {'@': 116}), 20: (1, {'@': 116}), 21: (1, {'@': 116}), 22: (1, {'@': 116}), 23: (1, {'@': 116}), 24: (1, {'@': 116}), 25: (1, {'@': 116}), 26: (1, {'@': 116}), 27: (1, {'@': 116}), 28: (1, {'@': 116}), 29: (1, {'@': 116})}, 1: {12: (1, {'@': 172})}, 2: {30: (1, {'@': 56})}, 3: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 87), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 4: {0: (1, {'@': 101}), 1: (1, {'@': 101}), 2: (1, {'@': 101}), 3: (1, {'@': 101}), 4: (1, {'@': 101}), 5: (1, {'@': 101}), 6: (1, {'@': 101}), 7: (1, {'@': 101}), 8: (1, {'@': 101}), 9: (1, {'@': 101}), 10: (1, {'@': 101}), 11: (1, {'@': 101}), 12: (1, {'@': 101}), 13: (1, {'@': 101}), 14: (1, {'@': 101}), 15: (1, {'@': 101}), 16: (1, {'@': 101}), 17: (1, {'@': 101}), 18: (1, {'@': 101}), 19: (1, {'@': 101}), 20: (1, {'@': 101}), 21: (1, {'@': 101}), 22: (1, {'@': 101}), 23: (1, {'@': 101}), 24: (1, {'@': 101}), 25: (1, {'@': 101}), 26: (1, {'@': 101}), 27: (1, {'@': 101}), 28: (1, {'@': 101}), 29: (1, {'@': 101})}, 5: {13: (0, 160), 0: (1, {'@': 165}), 19: (1, {'@': 165}), 24: (1, {'@': 165}), 8: (1, {'@': 165}), 4: (1, {'@': 165}), 18: (1, {'@': 165}), 29: (1, {'@': 165}), 12: (1, {'@': 165}), 6: (1, {'@': 165}), 26: (1, {'@': 165})}, 6: {30: (1, {'@': 58})}, 7: {0: (1, {'@': 134}), 1: (1, {'@': 134}), 2: (1, {'@': 134}), 3: (1, {'@': 134}), 4: (1, {'@': 134}), 5: (1, {'@': 134}), 6: (1, {'@': 134}), 7: (1, {'@': 134}), 8: (1, {'@': 134}), 9: (1, {'@': 134}), 10: (1, {'@': 134}), 11: (1, {'@': 134}), 12: (1, {'@': 134}), 13: (1, {'@': 134}), 14: (1, {'@': 134}), 15: (1, {'@': 134}), 16: (1, {'@': 134}), 17: (1, {'@': 134}), 18: (1, {'@': 134}), 19: (1, {'@': 134}), 20: (1, {'@': 134}), 21: (1, {'@': 134}), 22: (1, {'@': 134}), 23: (1, {'@': 134}), 24: (1, {'@': 134}), 25: (1, {'@': 134}), 26: (1, {'@': 134}), 27: (1, {'@': 134}), 28: (1, {'@': 134}), 29: (1, {'@': 134})}, 8: {0: (1, {'@': 139}), 1: (1, {'@': 139}), 2: (1, {'@': 139}), 3: (1, {'@': 139}), 4: (1, {'@': 139}), 5: (1, {'@': 139}), 6: (1, {'@': 139}), 7: (1, {'@': 139}), 8: (1, {'@': 139}), 9: (1, {'@': 139}), 10: (1, {'@': 139}), 11: (1, {'@': 139}), 12: (1, {'@': 139}), 13: (1, {'@': 139}), 14: (1, {'@': 139}), 15: (1, {'@': 139}), 16: (1, {'@': 139}), 17: (1, {'@': 139}), 18: (1, {'@': 139}), 19: (1, {'@': 139}), 20: (1, {'@': 139}), 21: (1, {'@': 139}), 22: (1, {'@': 139}), 23: (1, {'@': 139}), 24: (1, {'@': 139}), 25: (1, {'@': 139}), 26: (1, {'@': 139}), 27: (1, {'@': 139}), 28: (1, {'@': 139}), 29: (1, {'@': 139})}, 9: {52: (0, 131), 48: (0, 48), 39: (0, 226), 43: (0, 119), 57: (0, 236)}, 10: {63: (0, 157), 12: (0, 108), 17: (0, 166)}, 11: {0: (1, {'@': 106}), 1: (1, {'@': 106}), 2: (1, {'@': 106}), 3: (1, {'@': 106}), 4: (1, {'@': 106}), 5: (1, {'@': 106}), 6: (1, {'@': 106}), 7: (1, {'@': 106}), 8: (1, {'@': 106}), 9: (1, {'@': 106}), 10: (1, {'@': 106}), 11: (1, {'@': 106}), 12: (1, {'@': 106}), 13: (1, {'@': 106}), 14: (1, {'@': 106}), 15: (1, {'@': 106}), 16: (1, {'@': 106}), 17: (1, {'@': 106}), 18: (1, {'@': 106}), 19: (1, {'@': 106}), 20: (1, {'@': 106}), 21: (1, {'@': 106}), 22: (1, {'@': 106}), 23: (1, {'@': 106}), 24: (1, {'@': 106}), 25: (1, {'@': 106}), 26: (1, {'@': 106}), 27: (1, {'@': 106}), 28: (1, {'@': 106}), 29: (1, {'@': 106})}, 12: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 168), 38: (0, 76), 39: (0, 19), 40: (0, 17), 65: (0, 203), 66: (0, 222), 67: (0, 169), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 69: (0, 161), 70: (0, 158), 71: (0, 177), 59: (0, 0), 72: (0, 80), 73: (0, 97), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11), 74: (0, 124), 75: (0, 41), 35: (0, 35), 37: (0, 32), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 13: {0: (1, {'@': 117}), 1: (1, {'@': 117}), 2: (1, {'@': 117}), 3: (1, {'@': 117}), 4: (1, {'@': 117}), 5: (1, {'@': 117}), 6: (1, {'@': 117}), 7: (1, {'@': 117}), 8: (1, {'@': 117}), 9: (1, {'@': 117}), 10: (1, {'@': 117}), 11: (1, {'@': 117}), 12: (1, {'@': 117}), 13: (1, {'@': 117}), 14: (1, {'@': 117}), 15: (1, {'@': 117}), 16: (1, {'@': 117}), 17: (1, {'@': 117}), 18: (1, {'@': 117}), 19: (1, {'@': 117}), 20: (1, {'@': 117}), 21: (1, {'@': 117}), 22: (1, {'@': 117}), 23: (1, {'@': 117}), 24: (1, {'@': 117}), 25: (1, {'@': 117}), 26: (1, {'@': 117}), 27: (1, {'@': 117}), 28: (1, {'@': 117}), 29: (1, {'@': 117})}, 14: {0: (1, {'@': 135}), 1: (1, {'@': 135}), 2: (1, {'@': 135}), 3: (1, {'@': 135}), 4: (1, {'@': 135}), 5: (1, {'@': 135}), 6: (1, {'@': 135}), 7: (1, {'@': 135}), 8: (1, {'@': 135}), 9: (1, {'@': 135}), 10: (1, {'@': 135}), 11: (1, {'@': 135}), 12: (1, {'@': 135}), 13: (1, {'@': 135}), 14: (1, {'@': 135}), 15: (1, {'@': 135}), 16: (1, {'@': 135}), 17: (1, {'@': 135}), 18: (1, {'@': 135}), 19: (1, {'@': 135}), 20: (1, {'@': 135}), 21: (1, {'@': 135}), 22: (1, {'@': 135}), 23: (1, {'@': 135}), 24: (1, {'@': 135}), 25: (1, {'@': 135}), 26: (1, {'@': 135}), 27: (1, {'@': 135}), 28: (1, {'@': 135}), 29: (1, {'@': 135})}, 15: {5: (0, 57), 1: (0, 66), 28: (0, 73), 15: (0, 81)}, 16: {0: (1, {'@': 211}), 12: (1, {'@': 211}), 6: (1, {'@': 211}), 19: (1, {'@': 211}), 24: (1, {'@': 211}), 8: (1, {'@': 211}), 26: (1, {'@': 211})}, 17: {0: (1, {'@': 104}), 1: (1, {'@': 104}), 2: (1, {'@': 104}), 3: (1, {'@': 104}), 4: (1, {'@': 104}), 5: (1, {'@': 104}), 6: (1, {'@': 104}), 7: (1, {'@': 104}), 8: (1, {'@': 104}), 9: (1, {'@': 104}), 10: (1, {'@': 104}), 11: (1, {'@': 104}), 12: (1, {'@': 104}), 13: (1, {'@': 104}), 14: (1, {'@': 104}), 15: (1, {'@': 104}), 16: (1, {'@': 104}), 17: (1, {'@': 104}), 18: (1, {'@': 104}), 19: (1, {'@': 104}), 20: (1, {'@': 104}), 21: (1, {'@': 104}), 22: (1, {'@': 104}), 23: (1, {'@': 104}), 24: (1, {'@': 104}), 25: (1, {'@': 104}), 26: (1, {'@': 104}), 27: (1, {'@': 104}), 28: (1, {'@': 104}), 29: (1, {'@': 104})}, 18: {0: (1, {'@': 107}), 1: (1, {'@': 107}), 2: (1, {'@': 107}), 3: (1, {'@': 107}), 4: (1, {'@': 107}), 5: (1, {'@': 107}), 6: (1, {'@': 107}), 7: (1, {'@': 107}), 8: (1, {'@': 107}), 9: (1, {'@': 107}), 10: (1, {'@': 107}), 11: (1, {'@': 107}), 12: (1, {'@': 107}), 13: (1, {'@': 107}), 14: (1, {'@': 107}), 15: (1, {'@': 107}), 16: (1, {'@': 107}), 17: (1, {'@': 107}), 18: (1, {'@': 107}), 19: (1, {'@': 107}), 20: (1, {'@': 107}), 21: (1, {'@': 107}), 22: (1, {'@': 107}), 23: (1, {'@': 107}), 24: (1, {'@': 107}), 25: (1, {'@': 107}), 26: (1, {'@': 107}), 27: (1, {'@': 107}), 28: (1, {'@': 107}), 29: (1, {'@': 107})}, 19: {33: (0, 218), 0: (1, {'@': 103}), 1: (1, {'@': 103}), 2: (1, {'@': 103}), 3: (1, {'@': 103}), 4: (1, {'@': 103}), 5: (1, {'@': 103}), 6: (1, {'@': 103}), 7: (1, {'@': 103}), 8: (1, {'@': 103}), 9: (1, {'@': 103}), 10: (1, {'@': 103}), 11: (1, {'@': 103}), 12: (1, {'@': 103}), 13: (1, {'@': 103}), 14: (1, {'@': 103}), 15: (1, {'@': 103}), 16: (1, {'@': 103}), 17: (1, {'@': 103}), 18: (1, {'@': 103}), 19: (1, {'@': 103}), 20: (1, {'@': 103}), 21: (1, {'@': 103}), 22: (1, {'@': 103}), 23: (1, {'@': 103}), 24: (1, {'@': 103}), 25: (1, {'@': 103}), 26: (1, {'@': 103}), 27: (1, {'@': 103}), 28: (1, {'@': 103}), 29: (1, {'@': 103})}, 20: {15: (1, {'@': 97}), 56: (1, {'@': 97}), 61: (1, {'@': 97}), 52: (1, {'@': 97}), 54: (1, {'@': 97}), 47: (1, {'@': 97}), 33: (1, {'@': 97}), 57: (1, {'@': 97}), 58: (1, {'@': 97}), 41: (1, {'@': 97}), 49: (1, {'@': 97}), 45: (1, {'@': 97})}, 21: {78: (0, 130), 54: (0, 141)}, 22: {30: (1, {'@': 189})}, 23: {15: (1, {'@': 177}), 56: (1, {'@': 177}), 61: (1, {'@': 177}), 52: (1, {'@': 177}), 54: (1, {'@': 177}), 47: (1, {'@': 177}), 33: (1, {'@': 177}), 57: (1, {'@': 177}), 58: (1, {'@': 177}), 41: (1, {'@': 177}), 49: (1, {'@': 177}), 45: (1, {'@': 177}), 12: (1, {'@': 177})}, 24: {12: (1, {'@': 174}), 15: (1, {'@': 174}), 56: (1, {'@': 174}), 61: (1, {'@': 174}), 52: (1, {'@': 174}), 54: (1, {'@': 174}), 47: (1, {'@': 174}), 33: (1, {'@': 174}), 57: (1, {'@': 174}), 58: (1, {'@': 174}), 41: (1, {'@': 174}), 49: (1, {'@': 174}), 45: (1, {'@': 174})}, 25: {79: (0, 241), 80: (0, 250), 81: (0, 257), 82: (0, 128), 83: (0, 139), 84: (0, 146), 85: (0, 154), 30: (1, {'@': 65})}, 26: {24: (0, 189), 12: (0, 137)}, 27: {15: (1, {'@': 176}), 56: (1, {'@': 176}), 61: (1, {'@': 176}), 52: (1, {'@': 176}), 54: (1, {'@': 176}), 47: (1, {'@': 176}), 33: (1, {'@': 176}), 57: (1, {'@': 176}), 58: (1, {'@': 176}), 41: (1, {'@': 176}), 49: (1, {'@': 176}), 45: (1, {'@': 176})}, 28: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 175), 38: (0, 76), 39: (0, 19), 40: (0, 17), 65: (0, 203), 66: (0, 222), 67: (0, 169), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 71: (0, 184), 50: (0, 37), 69: (0, 161), 70: (0, 158), 59: (0, 0), 72: (0, 80), 73: (0, 97), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11), 74: (0, 124), 75: (0, 41), 35: (0, 35), 37: (0, 32), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 29: {15: (1, {'@': 89}), 56: (1, {'@': 89}), 61: (1, {'@': 89}), 52: (1, {'@': 89}), 54: (1, {'@': 89}), 47: (1, {'@': 89}), 33: (1, {'@': 89}), 57: (1, {'@': 89}), 58: (1, {'@': 89}), 41: (1, {'@': 89}), 49: (1, {'@': 89}), 45: (1, {'@': 89})}, 30: {54: (1, {'@': 197}), 78: (1, {'@': 197})}, 31: {86: (0, 47), 87: (0, 56)}, 32: {0: (1, {'@': 110}), 1: (1, {'@': 110}), 2: (1, {'@': 110}), 3: (1, {'@': 110}), 4: (1, {'@': 110}), 5: (1, {'@': 110}), 6: (1, {'@': 110}), 7: (1, {'@': 110}), 8: (1, {'@': 110}), 9: (1, {'@': 110}), 10: (1, {'@': 110}), 11: (1, {'@': 110}), 12: (1, {'@': 110}), 13: (1, {'@': 110}), 14: (1, {'@': 110}), 15: (1, {'@': 110}), 16: (1, {'@': 110}), 17: (1, {'@': 110}), 18: (1, {'@': 110}), 19: (1, {'@': 110}), 20: (1, {'@': 110}), 21: (1, {'@': 110}), 22: (1, {'@': 110}), 23: (1, {'@': 110}), 24: (1, {'@': 110}), 25: (1, {'@': 110}), 26: (1, {'@': 110}), 27: (1, {'@': 110}), 28: (1, {'@': 110}), 29: (1, {'@': 110})}, 33: {79: (0, 241), 83: (0, 162), 88: (0, 171), 82: (0, 128), 84: (0, 146), 81: (0, 179), 85: (0, 154), 80: (0, 250), 30: (1, {'@': 77})}, 34: {0: (1, {'@': 124}), 1: (1, {'@': 124}), 2: (1, {'@': 124}), 3: (1, {'@': 124}), 4: (1, {'@': 124}), 5: (1, {'@': 124}), 6: (1, {'@': 124}), 7: (1, {'@': 124}), 8: (1, {'@': 124}), 9: (1, {'@': 124}), 10: (1, {'@': 124}), 11: (1, {'@': 124}), 12: (1, {'@': 124}), 13: (1, {'@': 124}), 14: (1, {'@': 124}), 15: (1, {'@': 124}), 16: (1, {'@': 124}), 17: (1, {'@': 124}), 18: (1, {'@': 124}), 19: (1, {'@': 124}), 20: (1, {'@': 124}), 21: (1, {'@': 124}), 22: (1, {'@': 124}), 23: (1, {'@': 124}), 24: (1, {'@': 124}), 25: (1, {'@': 124}), 26: (1, {'@': 124}), 27: (1, {'@': 124}), 28: (1, {'@': 124}), 29: (1, {'@': 124})}, 35: {0: (1, {'@': 109}), 1: (1, {'@': 109}), 2: (1, {'@': 109}), 3: (1, {'@': 109}), 4: (1, {'@': 109}), 5: (1, {'@': 109}), 6: (1, {'@': 109}), 7: (1, {'@': 109}), 8: (1, {'@': 109}), 9: (1, {'@': 109}), 10: (1, {'@': 109}), 11: (1, {'@': 109}), 12: (1, {'@': 109}), 13: (1, {'@': 109}), 14: (1, {'@': 109}), 15: (1, {'@': 109}), 16: (1, {'@': 109}), 17: (1, {'@': 109}), 18: (1, {'@': 109}), 19: (1, {'@': 109}), 20: (1, {'@': 109}), 21: (1, {'@': 109}), 22: (1, {'@': 109}), 23: (1, {'@': 109}), 24: (1, {'@': 109}), 25: (1, {'@': 109}), 26: (1, {'@': 109}), 27: (1, {'@': 109}), 28: (1, {'@': 109}), 29: (1, {'@': 109})}, 36: {19: (0, 190), 0: (1, {'@': 150}), 24: (1, {'@': 150}), 12: (1, {'@': 150}), 26: (1, {'@': 150}), 6: (1, {'@': 150})}, 37: {0: (1, {'@': 112}), 1: (1, {'@': 112}), 2: (1, {'@': 112}), 3: (1, {'@': 112}), 4: (1, {'@': 112}), 5: (1, {'@': 112}), 6: (1, {'@': 112}), 7: (1, {'@': 112}), 8: (1, {'@': 112}), 9: (1, {'@': 112}), 10: (1, {'@': 112}), 11: (1, {'@': 112}), 12: (1, {'@': 112}), 13: (1, {'@': 112}), 14: (1, {'@': 112}), 15: (1, {'@': 112}), 16: (1, {'@': 112}), 17: (1, {'@': 112}), 18: (1, {'@': 112}), 19: (1, {'@': 112}), 20: (1, {'@': 112}), 21: (1, {'@': 112}), 22: (1, {'@': 112}), 23: (1, {'@': 112}), 24: (1, {'@': 112}), 25: (1, {'@': 112}), 26: (1, {'@': 112}), 27: (1, {'@': 112}), 28: (1, {'@': 112}), 29: (1, {'@': 112})}, 38: {15: (1, {'@': 93}), 56: (1, {'@': 93}), 61: (1, {'@': 93}), 52: (1, {'@': 93}), 54: (1, {'@': 93}), 47: (1, {'@': 93}), 33: (1, {'@': 93}), 57: (1, {'@': 93}), 58: (1, {'@': 93}), 41: (1, {'@': 93}), 49: (1, {'@': 93}), 45: (1, {'@': 93})}, 39: {23: (0, 148)}, 40: {56: (1, {'@': 191}), 61: (1, {'@': 191}), 52: (1, {'@': 191}), 47: (1, {'@': 191}), 33: (1, {'@': 191}), 57: (1, {'@': 191}), 58: (1, {'@': 191}), 49: (1, {'@': 191}), 45: (1, {'@': 191})}, 41: {89: (0, 185), 29: (0, 192), 0: (1, {'@': 160}), 8: (1, {'@': 160}), 4: (1, {'@': 160}), 18: (1, {'@': 160}), 19: (1, {'@': 160}), 12: (1, {'@': 160}), 6: (1, {'@': 160}), 24: (1, {'@': 160}), 26: (1, {'@': 160})}, 42: {81: (0, 187), 85: (0, 154), 80: (0, 250), 82: (0, 128), 84: (0, 146), 30: (1, {'@': 81})}, 43: {12: (0, 137), 24: (0, 196)}, 44: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 64: (0, 206), 48: (0, 48), 68: (0, 217), 50: (0, 37), 59: (0, 0), 72: (0, 80), 66: (0, 213), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 45: {15: (1, {'@': 88}), 56: (1, {'@': 88}), 61: (1, {'@': 88}), 52: (1, {'@': 88}), 54: (1, {'@': 88}), 47: (1, {'@': 88}), 33: (1, {'@': 88}), 57: (1, {'@': 88}), 58: (1, {'@': 88}), 41: (1, {'@': 88}), 49: (1, {'@': 88}), 45: (1, {'@': 88})}, 46: {0: (1, {'@': 119}), 1: (1, {'@': 119}), 2: (1, {'@': 119}), 3: (1, {'@': 119}), 4: (1, {'@': 119}), 5: (1, {'@': 119}), 6: (1, {'@': 119}), 7: (1, {'@': 119}), 8: (1, {'@': 119}), 9: (1, {'@': 119}), 10: (1, {'@': 119}), 11: (1, {'@': 119}), 12: (1, {'@': 119}), 13: (1, {'@': 119}), 14: (1, {'@': 119}), 15: (1, {'@': 119}), 16: (1, {'@': 119}), 17: (1, {'@': 119}), 18: (1, {'@': 119}), 19: (1, {'@': 119}), 20: (1, {'@': 119}), 21: (1, {'@': 119}), 22: (1, {'@': 119}), 23: (1, {'@': 119}), 24: (1, {'@': 119}), 25: (1, {'@': 119}), 26: (1, {'@': 119}), 27: (1, {'@': 119}), 28: (1, {'@': 119}), 29: (1, {'@': 119})}, 47: {79: (0, 65)}, 48: {0: (1, {'@': 194}), 1: (1, {'@': 194}), 2: (1, {'@': 194}), 3: (1, {'@': 194}), 4: (1, {'@': 194}), 5: (1, {'@': 194}), 6: (1, {'@': 194}), 7: (1, {'@': 194}), 8: (1, {'@': 194}), 9: (1, {'@': 194}), 10: (1, {'@': 194}), 11: (1, {'@': 194}), 12: (1, {'@': 194}), 13: (1, {'@': 194}), 14: (1, {'@': 194}), 15: (1, {'@': 194}), 16: (1, {'@': 194}), 17: (1, {'@': 194}), 18: (1, {'@': 194}), 19: (1, {'@': 194}), 20: (1, {'@': 194}), 21: (1, {'@': 194}), 22: (1, {'@': 194}), 23: (1, {'@': 194}), 24: (1, {'@': 194}), 33: (1, {'@': 194}), 25: (1, {'@': 194}), 27: (1, {'@': 194}), 28: (1, {'@': 194}), 26: (1, {'@': 194}), 29: (1, {'@': 194})}, 49: {32: (0, 125), 33: (0, 116), 90: (0, 61), 38: (0, 76), 39: (0, 19), 46: (0, 122), 40: (0, 17), 31: (0, 15), 42: (0, 3), 43: (0, 119), 44: (0, 96), 47: (0, 83), 48: (0, 48), 50: (0, 37), 54: (0, 59), 59: (0, 0), 61: (0, 27), 55: (0, 69), 62: (0, 11), 91: (0, 72), 35: (0, 35), 37: (0, 32), 45: (0, 52), 92: (0, 109), 49: (0, 23), 51: (0, 89), 52: (0, 131), 93: (0, 133), 53: (0, 151), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149)}, 50: {81: (0, 194), 85: (0, 154), 80: (0, 250), 82: (0, 128), 84: (0, 146), 30: (1, {'@': 69})}, 51: {0: (1, {'@': 126}), 1: (1, {'@': 126}), 2: (1, {'@': 126}), 3: (1, {'@': 126}), 4: (1, {'@': 126}), 5: (1, {'@': 126}), 6: (1, {'@': 126}), 7: (1, {'@': 126}), 8: (1, {'@': 126}), 9: (1, {'@': 126}), 10: (1, {'@': 126}), 11: (1, {'@': 126}), 12: (1, {'@': 126}), 13: (1, {'@': 126}), 14: (1, {'@': 126}), 15: (1, {'@': 126}), 16: (1, {'@': 126}), 17: (1, {'@': 126}), 18: (1, {'@': 126}), 19: (1, {'@': 126}), 20: (1, {'@': 126}), 21: (1, {'@': 126}), 22: (1, {'@': 126}), 23: (1, {'@': 126}), 24: (1, {'@': 126}), 25: (1, {'@': 126}), 26: (1, {'@': 126}), 27: (1, {'@': 126}), 28: (1, {'@': 126}), 29: (1, {'@': 126})}, 52: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 36: (0, 120), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 53: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 36: (0, 255), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 54: {15: (1, {'@': 95}), 56: (1, {'@': 95}), 61: (1, {'@': 95}), 52: (1, {'@': 95}), 54: (1, {'@': 95}), 47: (1, {'@': 95}), 33: (1, {'@': 95}), 57: (1, {'@': 95}), 58: (1, {'@': 95}), 41: (1, {'@': 95}), 49: (1, {'@': 95}), 45: (1, {'@': 95})}, 55: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 156), 38: (0, 76), 39: (0, 19), 65: (0, 203), 40: (0, 17), 66: (0, 222), 67: (0, 169), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 69: (0, 161), 70: (0, 158), 59: (0, 0), 72: (0, 80), 73: (0, 97), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11), 74: (0, 124), 75: (0, 41), 35: (0, 35), 37: (0, 32), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 71: (0, 164), 15: (0, 9)}, 56: {79: (1, {'@': 201})}, 57: {79: (0, 241), 81: (0, 200), 83: (0, 162), 82: (0, 128), 88: (0, 209), 84: (0, 146), 85: (0, 154), 80: (0, 250), 30: (1, {'@': 73})}, 58: {24: (0, 202), 12: (0, 137)}, 59: {94: (0, 106), 95: (0, 113)}, 60: {24: (0, 262), 12: (0, 137)}, 61: {30: (1, {'@': 57})}, 62: {12: (1, {'@': 169})}, 63: {24: (0, 186), 12: (0, 137)}, 64: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 173), 38: (0, 76), 39: (0, 19), 65: (0, 203), 40: (0, 17), 66: (0, 222), 67: (0, 169), 42: (0, 3), 43: (0, 119), 44: (0, 96), 71: (0, 181), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 69: (0, 161), 70: (0, 158), 59: (0, 0), 72: (0, 80), 73: (0, 97), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11), 74: (0, 124), 75: (0, 41), 35: (0, 35), 37: (0, 32), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 65: {30: (1, {'@': 188})}, 66: {79: (0, 241), 81: (0, 219), 83: (0, 228), 82: (0, 128), 84: (0, 146), 85: (0, 154), 80: (0, 250), 30: (1, {'@': 61})}, 67: {0: (1, {'@': 213}), 12: (1, {'@': 213}), 4: (1, {'@': 213}), 19: (1, {'@': 213}), 6: (1, {'@': 213}), 24: (1, {'@': 213}), 8: (1, {'@': 213}), 26: (1, {'@': 213})}, 68: {15: (1, {'@': 94}), 56: (1, {'@': 94}), 61: (1, {'@': 94}), 52: (1, {'@': 94}), 54: (1, {'@': 94}), 47: (1, {'@': 94}), 33: (1, {'@': 94}), 57: (1, {'@': 94}), 58: (1, {'@': 94}), 41: (1, {'@': 94}), 49: (1, {'@': 94}), 45: (1, {'@': 94})}, 69: {0: (1, {'@': 113}), 1: (1, {'@': 113}), 2: (1, {'@': 113}), 3: (1, {'@': 113}), 4: (1, {'@': 113}), 5: (1, {'@': 113}), 6: (1, {'@': 113}), 7: (1, {'@': 113}), 8: (1, {'@': 113}), 9: (1, {'@': 113}), 10: (1, {'@': 113}), 11: (1, {'@': 113}), 12: (1, {'@': 113}), 13: (1, {'@': 113}), 14: (1, {'@': 113}), 15: (1, {'@': 113}), 16: (1, {'@': 113}), 17: (1, {'@': 113}), 18: (1, {'@': 113}), 19: (1, {'@': 113}), 20: (1, {'@': 113}), 21: (1, {'@': 113}), 22: (1, {'@': 113}), 23: (1, {'@': 113}), 24: (1, {'@': 113}), 25: (1, {'@': 113}), 26: (1, {'@': 113}), 27: (1, {'@': 113}), 28: (1, {'@': 113}), 29: (1, {'@': 113})}, 70: {0: (1, {'@': 85}), 1: (1, {'@': 85}), 2: (1, {'@': 85}), 3: (1, {'@': 85}), 4: (1, {'@': 85}), 5: (1, {'@': 85}), 6: (1, {'@': 85}), 7: (1, {'@': 85}), 8: (1, {'@': 85}), 9: (1, {'@': 85}), 10: (1, {'@': 85}), 11: (1, {'@': 85}), 12: (1, {'@': 85}), 13: (1, {'@': 85}), 14: (1, {'@': 85}), 15: (1, {'@': 85}), 16: (1, {'@': 85}), 17: (1, {'@': 85}), 18: (1, {'@': 85}), 19: (1, {'@': 85}), 20: (1, {'@': 85}), 21: (1, {'@': 85}), 22: (1, {'@': 85}), 23: (1, {'@': 85}), 24: (1, {'@': 85}), 25: (1, {'@': 85}), 26: (1, {'@': 85}), 27: (1, {'@': 85}), 28: (1, {'@': 85}), 29: (1, {'@': 85})}, 71: {0: (1, {'@': 209}), 12: (1, {'@': 209}), 6: (1, {'@': 209}), 19: (1, {'@': 209}), 24: (1, {'@': 209}), 26: (1, {'@': 209})}, 72: {30: (1, {'@': 55})}, 73: {82: (0, 128), 81: (0, 237), 85: (0, 154), 80: (0, 250), 84: (0, 146), 30: (1, {'@': 79})}, 74: {0: (1, {'@': 215}), 12: (1, {'@': 215}), 4: (1, {'@': 215}), 18: (1, {'@': 215}), 19: (1, {'@': 215}), 6: (1, {'@': 215}), 24: (1, {'@': 215}), 8: (1, {'@': 215}), 26: (1, {'@': 215})}, 75: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 36: (0, 145), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 76: {0: (1, {'@': 114}), 1: (1, {'@': 114}), 2: (1, {'@': 114}), 3: (1, {'@': 114}), 4: (1, {'@': 114}), 5: (1, {'@': 114}), 6: (1, {'@': 114}), 7: (1, {'@': 114}), 8: (1, {'@': 114}), 9: (1, {'@': 114}), 10: (1, {'@': 114}), 11: (1, {'@': 114}), 12: (1, {'@': 114}), 13: (1, {'@': 114}), 14: (1, {'@': 114}), 15: (1, {'@': 114}), 16: (1, {'@': 114}), 17: (1, {'@': 114}), 18: (1, {'@': 114}), 19: (1, {'@': 114}), 20: (1, {'@': 114}), 21: (1, {'@': 114}), 22: (1, {'@': 114}), 23: (1, {'@': 114}), 24: (1, {'@': 114}), 25: (1, {'@': 114}), 26: (1, {'@': 114}), 27: (1, {'@': 114}), 28: (1, {'@': 114}), 29: (1, {'@': 114})}, 77: {12: (1, {'@': 167})}, 78: {0: (0, 193), 24: (1, {'@': 147}), 12: (1, {'@': 147}), 26: (1, {'@': 147}), 6: (1, {'@': 147})}, 79: {12: (1, {'@': 205}), 17: (1, {'@': 205}), 27: (1, {'@': 205}), 24: (1, {'@': 205})}, 80: {96: (0, 152), 4: (0, 159), 0: (1, {'@': 154}), 19: (1, {'@': 154}), 8: (1, {'@': 154}), 12: (1, {'@': 154}), 24: (1, {'@': 154}), 6: (1, {'@': 154}), 26: (1, {'@': 154})}, 81: {81: (0, 244), 85: (0, 154), 80: (0, 250), 82: (0, 128), 84: (0, 146), 30: (1, {'@': 67})}, 82: {0: (1, {'@': 217}), 4: (1, {'@': 217}), 6: (1, {'@': 217}), 8: (1, {'@': 217}), 12: (1, {'@': 217}), 18: (1, {'@': 217}), 19: (1, {'@': 217}), 24: (1, {'@': 217}), 26: (1, {'@': 217}), 29: (1, {'@': 217})}, 83: {97: (0, 101)}, 84: {15: (1, {'@': 98}), 56: (1, {'@': 98}), 61: (1, {'@': 98}), 52: (1, {'@': 98}), 54: (1, {'@': 98}), 47: (1, {'@': 98}), 33: (1, {'@': 98}), 57: (1, {'@': 98}), 58: (1, {'@': 98}), 41: (1, {'@': 98}), 49: (1, {'@': 98}), 45: (1, {'@': 98})}, 85: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 197), 67: (0, 199), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 59: (0, 0), 72: (0, 80), 66: (0, 213), 61: (0, 27), 55: (0, 69), 65: (0, 208), 31: (0, 4), 62: (0, 11), 74: (0, 124), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 86: {0: (1, {'@': 83}), 1: (1, {'@': 83}), 2: (1, {'@': 83}), 3: (1, {'@': 83}), 4: (1, {'@': 83}), 5: (1, {'@': 83}), 6: (1, {'@': 83}), 7: (1, {'@': 83}), 8: (1, {'@': 83}), 9: (1, {'@': 83}), 10: (1, {'@': 83}), 11: (1, {'@': 83}), 12: (1, {'@': 83}), 13: (1, {'@': 83}), 14: (1, {'@': 83}), 15: (1, {'@': 83}), 16: (1, {'@': 83}), 17: (1, {'@': 83}), 18: (1, {'@': 83}), 19: (1, {'@': 83}), 20: (1, {'@': 83}), 21: (1, {'@': 83}), 22: (1, {'@': 83}), 23: (1, {'@': 83}), 24: (1, {'@': 83}), 25: (1, {'@': 83}), 26: (1, {'@': 83}), 27: (1, {'@': 83}), 28: (1, {'@': 83}), 29: (1, {'@': 83})}, 87: {0: (1, {'@': 121}), 1: (1, {'@': 121}), 2: (1, {'@': 121}), 3: (1, {'@': 121}), 4: (1, {'@': 121}), 5: (1, {'@': 121}), 6: (1, {'@': 121}), 7: (1, {'@': 121}), 8: (1, {'@': 121}), 9: (1, {'@': 121}), 10: (1, {'@': 121}), 11: (1, {'@': 121}), 12: (1, {'@': 121}), 13: (1, {'@': 121}), 14: (1, {'@': 121}), 15: (1, {'@': 121}), 16: (1, {'@': 121}), 17: (1, {'@': 121}), 18: (1, {'@': 121}), 19: (1, {'@': 121}), 20: (1, {'@': 121}), 21: (1, {'@': 121}), 22: (1, {'@': 121}), 23: (1, {'@': 121}), 24: (1, {'@': 121}), 25: (1, {'@': 121}), 26: (1, {'@': 121}), 27: (1, {'@': 121}), 28: (1, {'@': 121}), 29: (1, {'@': 121})}, 88: {0: (1, {'@': 131}), 1: (1, {'@': 131}), 2: (1, {'@': 131}), 3: (1, {'@': 131}), 4: (1, {'@': 131}), 5: (1, {'@': 131}), 6: (1, {'@': 131}), 7: (1, {'@': 131}), 8: (1, {'@': 131}), 9: (1, {'@': 131}), 10: (1, {'@': 131}), 11: (1, {'@': 131}), 12: (1, {'@': 131}), 13: (1, {'@': 131}), 14: (1, {'@': 131}), 15: (1, {'@': 131}), 16: (1, {'@': 131}), 17: (1, {'@': 131}), 18: (1, {'@': 131}), 19: (1, {'@': 131}), 20: (1, {'@': 131}), 21: (1, {'@': 131}), 22: (1, {'@': 131}), 23: (1, {'@': 131}), 24: (1, {'@': 131}), 25: (1, {'@': 131}), 26: (1, {'@': 131}), 27: (1, {'@': 131}), 28: (1, {'@': 131}), 29: (1, {'@': 131})}, 89: {32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 36: (0, 10), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 90: {15: (1, {'@': 90}), 56: (1, {'@': 90}), 61: (1, {'@': 90}), 52: (1, {'@': 90}), 54: (1, {'@': 90}), 47: (1, {'@': 90}), 33: (1, {'@': 90}), 57: (1, {'@': 90}), 58: (1, {'@': 90}), 41: (1, {'@': 90}), 49: (1, {'@': 90}), 45: (1, {'@': 90})}, 91: {12: (1, {'@': 173}), 15: (1, {'@': 173}), 56: (1, {'@': 173}), 61: (1, {'@': 173}), 52: (1, {'@': 173}), 54: (1, {'@': 173}), 47: (1, {'@': 173}), 33: (1, {'@': 173}), 57: (1, {'@': 173}), 58: (1, {'@': 173}), 41: (1, {'@': 173}), 49: (1, {'@': 173}), 45: (1, {'@': 173})}, 92: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 36: (0, 215), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 93: {0: (1, {'@': 219}), 13: (1, {'@': 219}), 4: (1, {'@': 219}), 18: (1, {'@': 219}), 19: (1, {'@': 219}), 6: (1, {'@': 219}), 24: (1, {'@': 219}), 8: (1, {'@': 219}), 26: (1, {'@': 219}), 29: (1, {'@': 219}), 12: (1, {'@': 219})}, 94: {63: (0, 252), 12: (0, 108), 27: (0, 259)}, 95: {24: (0, 211), 12: (0, 137)}, 96: {32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 94), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 97: {24: (1, {'@': 146}), 12: (1, {'@': 146}), 26: (1, {'@': 146}), 6: (1, {'@': 146})}, 98: {15: (1, {'@': 92}), 56: (1, {'@': 92}), 61: (1, {'@': 92}), 52: (1, {'@': 92}), 54: (1, {'@': 92}), 47: (1, {'@': 92}), 33: (1, {'@': 92}), 57: (1, {'@': 92}), 58: (1, {'@': 92}), 41: (1, {'@': 92}), 49: (1, {'@': 92}), 45: (1, {'@': 92})}, 99: {0: (1, {'@': 128}), 1: (1, {'@': 128}), 2: (1, {'@': 128}), 3: (1, {'@': 128}), 4: (1, {'@': 128}), 5: (1, {'@': 128}), 6: (1, {'@': 128}), 7: (1, {'@': 128}), 8: (1, {'@': 128}), 9: (1, {'@': 128}), 10: (1, {'@': 128}), 11: (1, {'@': 128}), 12: (1, {'@': 128}), 13: (1, {'@': 128}), 14: (1, {'@': 128}), 15: (1, {'@': 128}), 16: (1, {'@': 128}), 17: (1, {'@': 128}), 18: (1, {'@': 128}), 19: (1, {'@': 128}), 20: (1, {'@': 128}), 21: (1, {'@': 128}), 22: (1, {'@': 128}), 23: (1, {'@': 128}), 24: (1, {'@': 128}), 25: (1, {'@': 128}), 26: (1, {'@': 128}), 27: (1, {'@': 128}), 28: (1, {'@': 128}), 29: (1, {'@': 128})}, 100: {0: (1, {'@': 136}), 1: (1, {'@': 136}), 2: (1, {'@': 136}), 3: (1, {'@': 136}), 4: (1, {'@': 136}), 5: (1, {'@': 136}), 6: (1, {'@': 136}), 7: (1, {'@': 136}), 8: (1, {'@': 136}), 9: (1, {'@': 136}), 10: (1, {'@': 136}), 11: (1, {'@': 136}), 12: (1, {'@': 136}), 13: (1, {'@': 136}), 14: (1, {'@': 136}), 15: (1, {'@': 136}), 16: (1, {'@': 136}), 17: (1, {'@': 136}), 18: (1, {'@': 136}), 19: (1, {'@': 136}), 20: (1, {'@': 136}), 21: (1, {'@': 136}), 22: (1, {'@': 136}), 23: (1, {'@': 136}), 24: (1, {'@': 136}), 25: (1, {'@': 136}), 26: (1, {'@': 136}), 27: (1, {'@': 136}), 28: (1, {'@': 136}), 29: (1, {'@': 136})}, 101: {0: (1, {'@': 108}), 1: (1, {'@': 108}), 2: (1, {'@': 108}), 3: (1, {'@': 108}), 4: (1, {'@': 108}), 5: (1, {'@': 108}), 6: (1, {'@': 108}), 7: (1, {'@': 108}), 8: (1, {'@': 108}), 9: (1, {'@': 108}), 10: (1, {'@': 108}), 11: (1, {'@': 108}), 12: (1, {'@': 108}), 13: (1, {'@': 108}), 14: (1, {'@': 108}), 15: (1, {'@': 108}), 16: (1, {'@': 108}), 17: (1, {'@': 108}), 18: (1, {'@': 108}), 19: (1, {'@': 108}), 20: (1, {'@': 108}), 21: (1, {'@': 108}), 22: (1, {'@': 108}), 23: (1, {'@': 108}), 24: (1, {'@': 108}), 25: (1, {'@': 108}), 26: (1, {'@': 108}), 27: (1, {'@': 108}), 28: (1, {'@': 108}), 29: (1, {'@': 108})}, 102: {0: (1, {'@': 86}), 1: (1, {'@': 86}), 2: (1, {'@': 86}), 3: (1, {'@': 86}), 4: (1, {'@': 86}), 5: (1, {'@': 86}), 6: (1, {'@': 86}), 7: (1, {'@': 86}), 8: (1, {'@': 86}), 9: (1, {'@': 86}), 10: (1, {'@': 86}), 11: (1, {'@': 86}), 12: (1, {'@': 86}), 13: (1, {'@': 86}), 14: (1, {'@': 86}), 15: (1, {'@': 86}), 16: (1, {'@': 86}), 17: (1, {'@': 86}), 18: (1, {'@': 86}), 19: (1, {'@': 86}), 20: (1, {'@': 86}), 21: (1, {'@': 86}), 22: (1, {'@': 86}), 23: (1, {'@': 86}), 24: (1, {'@': 86}), 25: (1, {'@': 86}), 26: (1, {'@': 86}), 27: (1, {'@': 86}), 28: (1, {'@': 86}), 29: (1, {'@': 86})}, 103: {15: (1, {'@': 96}), 56: (1, {'@': 96}), 61: (1, {'@': 96}), 52: (1, {'@': 96}), 54: (1, {'@': 96}), 47: (1, {'@': 96}), 33: (1, {'@': 96}), 57: (1, {'@': 96}), 58: (1, {'@': 96}), 41: (1, {'@': 96}), 49: (1, {'@': 96}), 45: (1, {'@': 96})}, 104: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 227), 38: (0, 76), 39: (0, 19), 65: (0, 203), 40: (0, 17), 66: (0, 222), 67: (0, 169), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 69: (0, 161), 70: (0, 158), 59: (0, 0), 72: (0, 80), 71: (0, 235), 73: (0, 97), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11), 74: (0, 124), 75: (0, 41), 35: (0, 35), 37: (0, 32), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 105: {0: (1, {'@': 137}), 1: (1, {'@': 137}), 2: (1, {'@': 137}), 3: (1, {'@': 137}), 4: (1, {'@': 137}), 5: (1, {'@': 137}), 6: (1, {'@': 137}), 7: (1, {'@': 137}), 8: (1, {'@': 137}), 9: (1, {'@': 137}), 10: (1, {'@': 137}), 11: (1, {'@': 137}), 12: (1, {'@': 137}), 13: (1, {'@': 137}), 14: (1, {'@': 137}), 15: (1, {'@': 137}), 16: (1, {'@': 137}), 17: (1, {'@': 137}), 18: (1, {'@': 137}), 19: (1, {'@': 137}), 20: (1, {'@': 137}), 21: (1, {'@': 137}), 22: (1, {'@': 137}), 23: (1, {'@': 137}), 24: (1, {'@': 137}), 25: (1, {'@': 137}), 26: (1, {'@': 137}), 27: (1, {'@': 137}), 28: (1, {'@': 137}), 29: (1, {'@': 137})}, 106: {78: (0, 132), 54: (0, 142)}, 107: {98: (0, 224), 99: (0, 230)}, 108: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 36: (0, 153), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 109: {32: (0, 125), 33: (0, 116), 38: (0, 76), 39: (0, 19), 46: (0, 122), 40: (0, 17), 31: (0, 15), 42: (0, 3), 43: (0, 119), 44: (0, 96), 47: (0, 83), 48: (0, 48), 50: (0, 37), 90: (0, 2), 59: (0, 0), 61: (0, 27), 55: (0, 69), 62: (0, 11), 35: (0, 35), 37: (0, 32), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149)}, 110: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 50: (0, 37), 59: (0, 0), 72: (0, 80), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 68: (0, 242), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 111: {0: (1, {'@': 207}), 12: (1, {'@': 207}), 24: (1, {'@': 207}), 26: (1, {'@': 207}), 6: (1, {'@': 207})}, 112: {24: (1, {'@': 142}), 0: (1, {'@': 158}), 4: (1, {'@': 158}), 19: (1, {'@': 158}), 8: (1, {'@': 158}), 12: (1, {'@': 142}), 26: (1, {'@': 142}), 6: (1, {'@': 142})}, 113: {54: (1, {'@': 196}), 78: (1, {'@': 196})}, 114: {30: (1, {'@': 190})}, 115: {15: (1, {'@': 99}), 56: (1, {'@': 99}), 61: (1, {'@': 99}), 52: (1, {'@': 99}), 54: (1, {'@': 99}), 47: (1, {'@': 99}), 33: (1, {'@': 99}), 57: (1, {'@': 99}), 58: (1, {'@': 99}), 41: (1, {'@': 99}), 49: (1, {'@': 99}), 45: (1, {'@': 99})}, 116: {32: (0, 125), 100: (0, 225), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 134), 62: (0, 174), 38: (0, 76), 39: (0, 19), 65: (0, 203), 40: (0, 17), 66: (0, 222), 67: (0, 169), 42: (0, 216), 43: (0, 119), 44: (0, 96), 46: (0, 248), 71: (0, 183), 47: (0, 83), 48: (0, 48), 68: (0, 217), 101: (0, 233), 50: (0, 37), 102: (0, 165), 69: (0, 161), 70: (0, 158), 103: (0, 204), 8: (0, 143), 104: (0, 135), 59: (0, 0), 72: (0, 80), 6: (0, 24), 73: (0, 97), 4: (0, 62), 61: (0, 27), 55: (0, 69), 31: (0, 4), 74: (0, 124), 75: (0, 41), 35: (0, 35), 26: (0, 91), 37: (0, 32), 19: (0, 77), 76: (0, 112), 41: (0, 178), 77: (0, 191), 45: (0, 52), 49: (0, 23), 29: (0, 205), 18: (0, 212), 105: (0, 243), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 106: (0, 249), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 13: (0, 256), 0: (0, 1), 15: (0, 9)}, 117: {8: (0, 251), 0: (1, {'@': 153}), 24: (1, {'@': 153}), 19: (1, {'@': 153}), 12: (1, {'@': 153}), 26: (1, {'@': 153}), 6: (1, {'@': 153})}, 118: {0: (1, {'@': 122}), 1: (1, {'@': 122}), 2: (1, {'@': 122}), 3: (1, {'@': 122}), 4: (1, {'@': 122}), 5: (1, {'@': 122}), 6: (1, {'@': 122}), 7: (1, {'@': 122}), 8: (1, {'@': 122}), 9: (1, {'@': 122}), 10: (1, {'@': 122}), 11: (1, {'@': 122}), 12: (1, {'@': 122}), 13: (1, {'@': 122}), 14: (1, {'@': 122}), 15: (1, {'@': 122}), 16: (1, {'@': 122}), 17: (1, {'@': 122}), 18: (1, {'@': 122}), 19: (1, {'@': 122}), 20: (1, {'@': 122}), 21: (1, {'@': 122}), 22: (1, {'@': 122}), 23: (1, {'@': 122}), 24: (1, {'@': 122}), 25: (1, {'@': 122}), 26: (1, {'@': 122}), 27: (1, {'@': 122}), 28: (1, {'@': 122}), 29: (1, {'@': 122})}, 119: {0: (1, {'@': 195}), 1: (1, {'@': 195}), 2: (1, {'@': 195}), 3: (1, {'@': 195}), 4: (1, {'@': 195}), 5: (1, {'@': 195}), 6: (1, {'@': 195}), 7: (1, {'@': 195}), 8: (1, {'@': 195}), 9: (1, {'@': 195}), 10: (1, {'@': 195}), 11: (1, {'@': 195}), 12: (1, {'@': 195}), 13: (1, {'@': 195}), 14: (1, {'@': 195}), 15: (1, {'@': 195}), 16: (1, {'@': 195}), 17: (1, {'@': 195}), 18: (1, {'@': 195}), 19: (1, {'@': 195}), 20: (1, {'@': 195}), 21: (1, {'@': 195}), 22: (1, {'@': 195}), 23: (1, {'@': 195}), 24: (1, {'@': 195}), 33: (1, {'@': 195}), 25: (1, {'@': 195}), 27: (1, {'@': 195}), 28: (1, {'@': 195}), 26: (1, {'@': 195}), 29: (1, {'@': 195})}, 120: {16: (0, 123), 2: (0, 20), 9: (0, 29), 14: (0, 38), 7: (0, 45), 10: (0, 68), 107: (0, 150), 25: (0, 84), 20: (0, 90), 21: (0, 98), 11: (0, 54), 22: (0, 103), 3: (0, 115)}, 121: {30: (1, {'@': 183}), 79: (1, {'@': 183})}, 122: {1: (0, 25), 5: (0, 33), 28: (0, 42), 15: (0, 50)}, 123: {15: (1, {'@': 91}), 56: (1, {'@': 91}), 61: (1, {'@': 91}), 52: (1, {'@': 91}), 54: (1, {'@': 91}), 47: (1, {'@': 91}), 33: (1, {'@': 91}), 57: (1, {'@': 91}), 58: (1, {'@': 91}), 41: (1, {'@': 91}), 49: (1, {'@': 91}), 45: (1, {'@': 91})}, 124: {108: (0, 167), 18: (0, 176), 0: (1, {'@': 157}), 4: (1, {'@': 157}), 19: (1, {'@': 157}), 8: (1, {'@': 157}), 12: (1, {'@': 157}), 6: (1, {'@': 157}), 24: (1, {'@': 157}), 26: (1, {'@': 157})}, 125: {0: (1, {'@': 102}), 1: (1, {'@': 102}), 2: (1, {'@': 102}), 3: (1, {'@': 102}), 4: (1, {'@': 102}), 5: (1, {'@': 102}), 6: (1, {'@': 102}), 7: (1, {'@': 102}), 8: (1, {'@': 102}), 9: (1, {'@': 102}), 10: (1, {'@': 102}), 11: (1, {'@': 102}), 12: (1, {'@': 102}), 13: (1, {'@': 102}), 14: (1, {'@': 102}), 15: (1, {'@': 102}), 16: (1, {'@': 102}), 17: (1, {'@': 102}), 18: (1, {'@': 102}), 19: (1, {'@': 102}), 20: (1, {'@': 102}), 21: (1, {'@': 102}), 22: (1, {'@': 102}), 23: (1, {'@': 102}), 24: (1, {'@': 102}), 25: (1, {'@': 102}), 26: (1, {'@': 102}), 27: (1, {'@': 102}), 28: (1, {'@': 102}), 29: (1, {'@': 102})}, 126: {32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 36: (0, 258), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 127: {0: (1, {'@': 138}), 1: (1, {'@': 138}), 2: (1, {'@': 138}), 3: (1, {'@': 138}), 4: (1, {'@': 138}), 5: (1, {'@': 138}), 6: (1, {'@': 138}), 7: (1, {'@': 138}), 8: (1, {'@': 138}), 9: (1, {'@': 138}), 10: (1, {'@': 138}), 11: (1, {'@': 138}), 12: (1, {'@': 138}), 13: (1, {'@': 138}), 14: (1, {'@': 138}), 15: (1, {'@': 138}), 16: (1, {'@': 138}), 17: (1, {'@': 138}), 18: (1, {'@': 138}), 19: (1, {'@': 138}), 20: (1, {'@': 138}), 21: (1, {'@': 138}), 22: (1, {'@': 138}), 23: (1, {'@': 138}), 24: (1, {'@': 138}), 25: (1, {'@': 138}), 26: (1, {'@': 138}), 27: (1, {'@': 138}), 28: (1, {'@': 138}), 29: (1, {'@': 138})}, 128: {30: (1, {'@': 186}), 79: (1, {'@': 186})}, 129: {12: (0, 108), 63: (0, 43), 24: (0, 51)}, 130: {95: (0, 239), 109: (0, 246)}, 131: {0: (1, {'@': 203}), 1: (1, {'@': 203}), 2: (1, {'@': 203}), 3: (1, {'@': 203}), 4: (1, {'@': 203}), 5: (1, {'@': 203}), 6: (1, {'@': 203}), 7: (1, {'@': 203}), 8: (1, {'@': 203}), 9: (1, {'@': 203}), 10: (1, {'@': 203}), 11: (1, {'@': 203}), 12: (1, {'@': 203}), 13: (1, {'@': 203}), 14: (1, {'@': 203}), 15: (1, {'@': 203}), 16: (1, {'@': 203}), 17: (1, {'@': 203}), 18: (1, {'@': 203}), 19: (1, {'@': 203}), 20: (1, {'@': 203}), 21: (1, {'@': 203}), 22: (1, {'@': 203}), 23: (1, {'@': 203}), 24: (1, {'@': 203}), 33: (1, {'@': 203}), 25: (1, {'@': 203}), 27: (1, {'@': 203}), 28: (1, {'@': 203}), 26: (1, {'@': 203}), 29: (1, {'@': 203})}, 132: {110: (0, 21), 99: (0, 30)}, 133: {}, 134: {13: (0, 53), 26: (0, 91), 63: (0, 60), 10: (0, 68), 107: (0, 75), 25: (0, 84), 20: (0, 90), 21: (0, 98), 22: (0, 103), 12: (0, 108), 3: (0, 115), 16: (0, 123), 111: (0, 5), 6: (0, 24), 100: (0, 12), 2: (0, 20), 9: (0, 29), 14: (0, 38), 7: (0, 45), 11: (0, 54), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 135: {12: (1, {'@': 179})}, 136: {32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 129), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 137: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 36: (0, 79), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 138: {112: (0, 36), 19: (0, 44), 0: (1, {'@': 148}), 12: (1, {'@': 148}), 24: (1, {'@': 148}), 26: (1, {'@': 148}), 6: (1, {'@': 148})}, 139: {30: (1, {'@': 64})}, 140: {12: (0, 108), 63: (0, 58)}, 141: {56: (1, {'@': 192}), 61: (1, {'@': 192}), 52: (1, {'@': 192}), 47: (1, {'@': 192}), 33: (1, {'@': 192}), 57: (1, {'@': 192}), 58: (1, {'@': 192}), 49: (1, {'@': 192}), 45: (1, {'@': 192})}, 142: {56: (1, {'@': 193}), 61: (1, {'@': 193}), 52: (1, {'@': 193}), 47: (1, {'@': 193}), 33: (1, {'@': 193}), 57: (1, {'@': 193}), 58: (1, {'@': 193}), 49: (1, {'@': 193}), 45: (1, {'@': 193})}, 143: {12: (1, {'@': 166})}, 144: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 36: (0, 140), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 145: {24: (0, 86)}, 146: {30: (1, {'@': 185}), 79: (1, {'@': 185})}, 147: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 50: (0, 37), 59: (0, 0), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 72: (0, 67), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 148: {0: (1, {'@': 82}), 1: (1, {'@': 82}), 2: (1, {'@': 82}), 3: (1, {'@': 82}), 4: (1, {'@': 82}), 5: (1, {'@': 82}), 6: (1, {'@': 82}), 7: (1, {'@': 82}), 8: (1, {'@': 82}), 9: (1, {'@': 82}), 10: (1, {'@': 82}), 11: (1, {'@': 82}), 12: (1, {'@': 82}), 13: (1, {'@': 82}), 14: (1, {'@': 82}), 15: (1, {'@': 82}), 16: (1, {'@': 82}), 17: (1, {'@': 82}), 18: (1, {'@': 82}), 19: (1, {'@': 82}), 20: (1, {'@': 82}), 21: (1, {'@': 82}), 22: (1, {'@': 82}), 23: (1, {'@': 82}), 24: (1, {'@': 82}), 25: (1, {'@': 82}), 26: (1, {'@': 82}), 27: (1, {'@': 82}), 28: (1, {'@': 82}), 29: (1, {'@': 82})}, 149: {0: (1, {'@': 115}), 1: (1, {'@': 115}), 2: (1, {'@': 115}), 3: (1, {'@': 115}), 4: (1, {'@': 115}), 5: (1, {'@': 115}), 6: (1, {'@': 115}), 7: (1, {'@': 115}), 8: (1, {'@': 115}), 9: (1, {'@': 115}), 10: (1, {'@': 115}), 11: (1, {'@': 115}), 12: (1, {'@': 115}), 13: (1, {'@': 115}), 14: (1, {'@': 115}), 15: (1, {'@': 115}), 16: (1, {'@': 115}), 17: (1, {'@': 115}), 18: (1, {'@': 115}), 19: (1, {'@': 115}), 20: (1, {'@': 115}), 21: (1, {'@': 115}), 22: (1, {'@': 115}), 23: (1, {'@': 115}), 24: (1, {'@': 115}), 25: (1, {'@': 115}), 26: (1, {'@': 115}), 27: (1, {'@': 115}), 28: (1, {'@': 115}), 29: (1, {'@': 115})}, 150: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 36: (0, 39), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 151: {0: (1, {'@': 111}), 1: (1, {'@': 111}), 2: (1, {'@': 111}), 3: (1, {'@': 111}), 4: (1, {'@': 111}), 5: (1, {'@': 111}), 6: (1, {'@': 111}), 7: (1, {'@': 111}), 8: (1, {'@': 111}), 9: (1, {'@': 111}), 10: (1, {'@': 111}), 11: (1, {'@': 111}), 12: (1, {'@': 111}), 13: (1, {'@': 111}), 14: (1, {'@': 111}), 15: (1, {'@': 111}), 16: (1, {'@': 111}), 17: (1, {'@': 111}), 18: (1, {'@': 111}), 19: (1, {'@': 111}), 20: (1, {'@': 111}), 21: (1, {'@': 111}), 22: (1, {'@': 111}), 23: (1, {'@': 111}), 24: (1, {'@': 111}), 25: (1, {'@': 111}), 26: (1, {'@': 111}), 27: (1, {'@': 111}), 28: (1, {'@': 111}), 29: (1, {'@': 111})}, 152: {4: (0, 147), 0: (1, {'@': 156}), 24: (1, {'@': 156}), 19: (1, {'@': 156}), 8: (1, {'@': 156}), 12: (1, {'@': 156}), 6: (1, {'@': 156}), 26: (1, {'@': 156})}, 153: {12: (1, {'@': 204}), 17: (1, {'@': 204}), 27: (1, {'@': 204}), 24: (1, {'@': 204})}, 154: {97: (0, 245)}, 155: {0: (1, {'@': 212}), 12: (1, {'@': 212}), 4: (1, {'@': 212}), 19: (1, {'@': 212}), 6: (1, {'@': 212}), 24: (1, {'@': 212}), 8: (1, {'@': 212}), 26: (1, {'@': 212})}, 156: {24: (0, 254), 13: (0, 53), 111: (0, 5), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 157: {17: (0, 46), 12: (0, 137)}, 158: {0: (1, {'@': 155}), 19: (1, {'@': 155}), 8: (1, {'@': 155}), 24: (1, {'@': 143}), 12: (1, {'@': 143}), 26: (1, {'@': 143}), 6: (1, {'@': 143})}, 159: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 50: (0, 37), 59: (0, 0), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 72: (0, 155), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 160: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 40: (0, 17), 41: (0, 178), 36: (0, 93), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 161: {24: (1, {'@': 140}), 0: (1, {'@': 164}), 19: (1, {'@': 164}), 8: (1, {'@': 164}), 4: (1, {'@': 164}), 18: (1, {'@': 164}), 29: (1, {'@': 164}), 12: (1, {'@': 140}), 26: (1, {'@': 140}), 6: (1, {'@': 140})}, 162: {30: (1, {'@': 187})}, 163: {31: (0, 4), 69: (0, 223), 75: (0, 41), 32: (0, 125), 34: (0, 221), 35: (0, 35), 33: (0, 116), 36: (0, 197), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 77: (0, 231), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 74: (0, 74), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 59: (0, 0), 58: (0, 182), 60: (0, 149), 61: (0, 27), 55: (0, 69), 62: (0, 11)}, 164: {24: (0, 261)}, 165: {12: (0, 136)}, 166: {0: (1, {'@': 120}), 1: (1, {'@': 120}), 2: (1, {'@': 120}), 3: (1, {'@': 120}), 4: (1, {'@': 120}), 5: (1, {'@': 120}), 6: (1, {'@': 120}), 7: (1, {'@': 120}), 8: (1, {'@': 120}), 9: (1, {'@': 120}), 10: (1, {'@': 120}), 11: (1, {'@': 120}), 12: (1, {'@': 120}), 13: (1, {'@': 120}), 14: (1, {'@': 120}), 15: (1, {'@': 120}), 16: (1, {'@': 120}), 17: (1, {'@': 120}), 18: (1, {'@': 120}), 19: (1, {'@': 120}), 20: (1, {'@': 120}), 21: (1, {'@': 120}), 22: (1, {'@': 120}), 23: (1, {'@': 120}), 24: (1, {'@': 120}), 25: (1, {'@': 120}), 26: (1, {'@': 120}), 27: (1, {'@': 120}), 28: (1, {'@': 120}), 29: (1, {'@': 120})}, 167: {18: (0, 163), 0: (1, {'@': 159}), 24: (1, {'@': 159}), 8: (1, {'@': 159}), 4: (1, {'@': 159}), 19: (1, {'@': 159}), 12: (1, {'@': 159}), 6: (1, {'@': 159}), 26: (1, {'@': 159})}, 168: {13: (0, 53), 111: (0, 5), 24: (0, 100), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 169: {113: (0, 78), 0: (0, 85)}, 170: {52: (0, 131), 48: (0, 48), 43: (0, 119), 39: (0, 214), 57: (0, 236)}, 171: {30: (1, {'@': 76})}, 172: {0: (1, {'@': 214}), 12: (1, {'@': 214}), 4: (1, {'@': 214}), 18: (1, {'@': 214}), 19: (1, {'@': 214}), 6: (1, {'@': 214}), 24: (1, {'@': 214}), 8: (1, {'@': 214}), 26: (1, {'@': 214})}, 173: {13: (0, 53), 111: (0, 5), 24: (0, 7), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 174: {63: (0, 63), 24: (0, 70), 12: (0, 108), 0: (1, {'@': 106}), 2: (1, {'@': 106}), 3: (1, {'@': 106}), 4: (1, {'@': 106}), 6: (1, {'@': 106}), 7: (1, {'@': 106}), 9: (1, {'@': 106}), 8: (1, {'@': 106}), 10: (1, {'@': 106}), 11: (1, {'@': 106}), 13: (1, {'@': 106}), 14: (1, {'@': 106}), 16: (1, {'@': 106}), 18: (1, {'@': 106}), 19: (1, {'@': 106}), 20: (1, {'@': 106}), 21: (1, {'@': 106}), 25: (1, {'@': 106}), 22: (1, {'@': 106}), 26: (1, {'@': 106}), 29: (1, {'@': 106})}, 175: {12: (0, 55), 13: (0, 53), 111: (0, 5), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 176: {31: (0, 4), 69: (0, 223), 75: (0, 41), 32: (0, 125), 34: (0, 221), 35: (0, 35), 33: (0, 116), 36: (0, 197), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 74: (0, 172), 42: (0, 3), 77: (0, 231), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 59: (0, 0), 58: (0, 182), 60: (0, 149), 61: (0, 27), 55: (0, 69), 62: (0, 11)}, 177: {24: (0, 105)}, 178: {52: (0, 131), 48: (0, 48), 43: (0, 119), 39: (0, 198), 57: (0, 236)}, 179: {79: (0, 241), 88: (0, 253), 83: (0, 162), 30: (1, {'@': 75})}, 180: {31: (0, 4), 69: (0, 223), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 197), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 75: (0, 82), 53: (0, 151), 54: (0, 170), 52: (0, 131), 56: (0, 232), 57: (0, 236), 59: (0, 0), 58: (0, 182), 60: (0, 149), 61: (0, 27), 55: (0, 69), 62: (0, 11)}, 181: {24: (0, 14)}, 182: {15: (1, {'@': 175}), 56: (1, {'@': 175}), 61: (1, {'@': 175}), 52: (1, {'@': 175}), 54: (1, {'@': 175}), 47: (1, {'@': 175}), 33: (1, {'@': 175}), 57: (1, {'@': 175}), 58: (1, {'@': 175}), 41: (1, {'@': 175}), 49: (1, {'@': 175}), 45: (1, {'@': 175})}, 183: {24: (0, 99), 100: (0, 104), 6: (0, 24), 26: (0, 91)}, 184: {12: (0, 64)}, 185: {29: (0, 180), 0: (1, {'@': 162}), 24: (1, {'@': 162}), 8: (1, {'@': 162}), 4: (1, {'@': 162}), 18: (1, {'@': 162}), 19: (1, {'@': 162}), 12: (1, {'@': 162}), 6: (1, {'@': 162}), 26: (1, {'@': 162})}, 186: {0: (1, {'@': 84}), 1: (1, {'@': 84}), 2: (1, {'@': 84}), 3: (1, {'@': 84}), 4: (1, {'@': 84}), 5: (1, {'@': 84}), 6: (1, {'@': 84}), 7: (1, {'@': 84}), 8: (1, {'@': 84}), 9: (1, {'@': 84}), 10: (1, {'@': 84}), 11: (1, {'@': 84}), 12: (1, {'@': 84}), 13: (1, {'@': 84}), 14: (1, {'@': 84}), 15: (1, {'@': 84}), 16: (1, {'@': 84}), 17: (1, {'@': 84}), 18: (1, {'@': 84}), 19: (1, {'@': 84}), 20: (1, {'@': 84}), 21: (1, {'@': 84}), 22: (1, {'@': 84}), 23: (1, {'@': 84}), 24: (1, {'@': 84}), 25: (1, {'@': 84}), 26: (1, {'@': 84}), 27: (1, {'@': 84}), 28: (1, {'@': 84}), 29: (1, {'@': 84})}, 187: {30: (1, {'@': 80})}, 188: {0: (1, {'@': 216}), 4: (1, {'@': 216}), 6: (1, {'@': 216}), 8: (1, {'@': 216}), 12: (1, {'@': 216}), 18: (1, {'@': 216}), 19: (1, {'@': 216}), 24: (1, {'@': 216}), 26: (1, {'@': 216}), 29: (1, {'@': 216})}, 189: {0: (1, {'@': 123}), 1: (1, {'@': 123}), 2: (1, {'@': 123}), 3: (1, {'@': 123}), 4: (1, {'@': 123}), 5: (1, {'@': 123}), 6: (1, {'@': 123}), 7: (1, {'@': 123}), 8: (1, {'@': 123}), 9: (1, {'@': 123}), 10: (1, {'@': 123}), 11: (1, {'@': 123}), 12: (1, {'@': 123}), 13: (1, {'@': 123}), 14: (1, {'@': 123}), 15: (1, {'@': 123}), 16: (1, {'@': 123}), 17: (1, {'@': 123}), 18: (1, {'@': 123}), 19: (1, {'@': 123}), 20: (1, {'@': 123}), 21: (1, {'@': 123}), 22: (1, {'@': 123}), 23: (1, {'@': 123}), 24: (1, {'@': 123}), 25: (1, {'@': 123}), 26: (1, {'@': 123}), 27: (1, {'@': 123}), 28: (1, {'@': 123}), 29: (1, {'@': 123})}, 190: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 64: (0, 71), 68: (0, 217), 50: (0, 37), 59: (0, 0), 72: (0, 80), 66: (0, 213), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 191: {24: (1, {'@': 141}), 0: (1, {'@': 161}), 8: (1, {'@': 161}), 4: (1, {'@': 161}), 18: (1, {'@': 161}), 19: (1, {'@': 161}), 12: (1, {'@': 141}), 26: (1, {'@': 141}), 6: (1, {'@': 141})}, 192: {31: (0, 4), 69: (0, 223), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 197), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 75: (0, 188), 56: (0, 232), 57: (0, 236), 59: (0, 0), 58: (0, 182), 60: (0, 149), 61: (0, 27), 55: (0, 69), 62: (0, 11)}, 193: {32: (0, 125), 33: (0, 116), 34: (0, 221), 64: (0, 138), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 68: (0, 217), 50: (0, 37), 59: (0, 0), 72: (0, 80), 66: (0, 213), 61: (0, 27), 55: (0, 69), 65: (0, 208), 31: (0, 4), 67: (0, 111), 62: (0, 11), 74: (0, 124), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 194: {30: (1, {'@': 68})}, 195: {24: (0, 88), 12: (0, 108), 63: (0, 95)}, 196: {0: (1, {'@': 125}), 1: (1, {'@': 125}), 2: (1, {'@': 125}), 3: (1, {'@': 125}), 4: (1, {'@': 125}), 5: (1, {'@': 125}), 6: (1, {'@': 125}), 7: (1, {'@': 125}), 8: (1, {'@': 125}), 9: (1, {'@': 125}), 10: (1, {'@': 125}), 11: (1, {'@': 125}), 12: (1, {'@': 125}), 13: (1, {'@': 125}), 14: (1, {'@': 125}), 15: (1, {'@': 125}), 16: (1, {'@': 125}), 17: (1, {'@': 125}), 18: (1, {'@': 125}), 19: (1, {'@': 125}), 20: (1, {'@': 125}), 21: (1, {'@': 125}), 22: (1, {'@': 125}), 23: (1, {'@': 125}), 24: (1, {'@': 125}), 25: (1, {'@': 125}), 26: (1, {'@': 125}), 27: (1, {'@': 125}), 28: (1, {'@': 125}), 29: (1, {'@': 125})}, 197: {13: (0, 53), 111: (0, 5), 0: (1, {'@': 163}), 12: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 6: (1, {'@': 163}), 19: (1, {'@': 163}), 24: (1, {'@': 163}), 8: (1, {'@': 163}), 26: (1, {'@': 163}), 29: (1, {'@': 163})}, 198: {0: (1, {'@': 181}), 1: (1, {'@': 181}), 2: (1, {'@': 181}), 3: (1, {'@': 181}), 4: (1, {'@': 181}), 5: (1, {'@': 181}), 6: (1, {'@': 181}), 7: (1, {'@': 181}), 8: (1, {'@': 181}), 9: (1, {'@': 181}), 10: (1, {'@': 181}), 11: (1, {'@': 181}), 12: (1, {'@': 181}), 13: (1, {'@': 181}), 14: (1, {'@': 181}), 15: (1, {'@': 181}), 16: (1, {'@': 181}), 17: (1, {'@': 181}), 18: (1, {'@': 181}), 19: (1, {'@': 181}), 20: (1, {'@': 181}), 21: (1, {'@': 181}), 22: (1, {'@': 181}), 23: (1, {'@': 181}), 24: (1, {'@': 181}), 25: (1, {'@': 181}), 26: (1, {'@': 181}), 27: (1, {'@': 181}), 28: (1, {'@': 181}), 29: (1, {'@': 181})}, 199: {0: (1, {'@': 206}), 12: (1, {'@': 206}), 24: (1, {'@': 206}), 26: (1, {'@': 206}), 6: (1, {'@': 206})}, 200: {79: (0, 241), 88: (0, 260), 83: (0, 162), 30: (1, {'@': 71})}, 201: {24: (0, 102), 12: (0, 137)}, 202: {0: (1, {'@': 127}), 1: (1, {'@': 127}), 2: (1, {'@': 127}), 3: (1, {'@': 127}), 4: (1, {'@': 127}), 5: (1, {'@': 127}), 6: (1, {'@': 127}), 7: (1, {'@': 127}), 8: (1, {'@': 127}), 9: (1, {'@': 127}), 10: (1, {'@': 127}), 11: (1, {'@': 127}), 12: (1, {'@': 127}), 13: (1, {'@': 127}), 14: (1, {'@': 127}), 15: (1, {'@': 127}), 16: (1, {'@': 127}), 17: (1, {'@': 127}), 18: (1, {'@': 127}), 19: (1, {'@': 127}), 20: (1, {'@': 127}), 21: (1, {'@': 127}), 22: (1, {'@': 127}), 23: (1, {'@': 127}), 24: (1, {'@': 127}), 25: (1, {'@': 127}), 26: (1, {'@': 127}), 27: (1, {'@': 127}), 28: (1, {'@': 127}), 29: (1, {'@': 127})}, 203: {0: (1, {'@': 149}), 24: (1, {'@': 145}), 12: (1, {'@': 145}), 26: (1, {'@': 145}), 6: (1, {'@': 145})}, 204: {12: (0, 144)}, 205: {12: (1, {'@': 170})}, 206: {0: (1, {'@': 208}), 12: (1, {'@': 208}), 6: (1, {'@': 208}), 19: (1, {'@': 208}), 24: (1, {'@': 208}), 26: (1, {'@': 208})}, 207: {36: (0, 195), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 208: {0: (1, {'@': 149}), 12: (1, {'@': 149}), 24: (1, {'@': 149}), 26: (1, {'@': 149}), 6: (1, {'@': 149})}, 209: {30: (1, {'@': 72})}, 210: {0: (1, {'@': 87}), 1: (1, {'@': 87}), 2: (1, {'@': 87}), 3: (1, {'@': 87}), 4: (1, {'@': 87}), 5: (1, {'@': 87}), 6: (1, {'@': 87}), 7: (1, {'@': 87}), 8: (1, {'@': 87}), 9: (1, {'@': 87}), 10: (1, {'@': 87}), 11: (1, {'@': 87}), 12: (1, {'@': 87}), 13: (1, {'@': 87}), 14: (1, {'@': 87}), 15: (1, {'@': 87}), 16: (1, {'@': 87}), 17: (1, {'@': 87}), 18: (1, {'@': 87}), 19: (1, {'@': 87}), 20: (1, {'@': 87}), 21: (1, {'@': 87}), 22: (1, {'@': 87}), 23: (1, {'@': 87}), 24: (1, {'@': 87}), 25: (1, {'@': 87}), 26: (1, {'@': 87}), 27: (1, {'@': 87}), 28: (1, {'@': 87}), 29: (1, {'@': 87})}, 211: {0: (1, {'@': 130}), 1: (1, {'@': 130}), 2: (1, {'@': 130}), 3: (1, {'@': 130}), 4: (1, {'@': 130}), 5: (1, {'@': 130}), 6: (1, {'@': 130}), 7: (1, {'@': 130}), 8: (1, {'@': 130}), 9: (1, {'@': 130}), 10: (1, {'@': 130}), 11: (1, {'@': 130}), 12: (1, {'@': 130}), 13: (1, {'@': 130}), 14: (1, {'@': 130}), 15: (1, {'@': 130}), 16: (1, {'@': 130}), 17: (1, {'@': 130}), 18: (1, {'@': 130}), 19: (1, {'@': 130}), 20: (1, {'@': 130}), 21: (1, {'@': 130}), 22: (1, {'@': 130}), 23: (1, {'@': 130}), 24: (1, {'@': 130}), 25: (1, {'@': 130}), 26: (1, {'@': 130}), 27: (1, {'@': 130}), 28: (1, {'@': 130}), 29: (1, {'@': 130})}, 212: {12: (1, {'@': 168})}, 213: {0: (1, {'@': 152}), 12: (1, {'@': 152}), 19: (1, {'@': 152}), 24: (1, {'@': 152}), 26: (1, {'@': 152}), 6: (1, {'@': 152})}, 214: {0: (1, {'@': 180}), 1: (1, {'@': 180}), 2: (1, {'@': 180}), 3: (1, {'@': 180}), 4: (1, {'@': 180}), 5: (1, {'@': 180}), 6: (1, {'@': 180}), 7: (1, {'@': 180}), 8: (1, {'@': 180}), 9: (1, {'@': 180}), 10: (1, {'@': 180}), 11: (1, {'@': 180}), 12: (1, {'@': 180}), 13: (1, {'@': 180}), 14: (1, {'@': 180}), 15: (1, {'@': 180}), 16: (1, {'@': 180}), 17: (1, {'@': 180}), 18: (1, {'@': 180}), 19: (1, {'@': 180}), 20: (1, {'@': 180}), 21: (1, {'@': 180}), 22: (1, {'@': 180}), 23: (1, {'@': 180}), 24: (1, {'@': 180}), 25: (1, {'@': 180}), 26: (1, {'@': 180}), 27: (1, {'@': 180}), 28: (1, {'@': 180}), 29: (1, {'@': 180})}, 215: {24: (0, 118)}, 216: {31: (0, 4), 32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 36: (0, 87), 37: (0, 32), 38: (0, 76), 39: (0, 19), 12: (0, 92), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 55: (0, 69), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 15: (0, 9), 62: (0, 11)}, 217: {8: (0, 110), 114: (0, 117), 0: (1, {'@': 151}), 19: (1, {'@': 151}), 12: (1, {'@': 151}), 24: (1, {'@': 151}), 26: (1, {'@': 151}), 6: (1, {'@': 151})}, 218: {32: (0, 125), 33: (0, 116), 34: (0, 221), 35: (0, 35), 37: (0, 32), 38: (0, 76), 39: (0, 19), 15: (0, 9), 40: (0, 17), 41: (0, 178), 42: (0, 3), 43: (0, 119), 44: (0, 96), 45: (0, 52), 46: (0, 248), 47: (0, 83), 48: (0, 48), 49: (0, 23), 36: (0, 234), 50: (0, 37), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 56: (0, 232), 57: (0, 236), 58: (0, 182), 59: (0, 0), 60: (0, 149), 61: (0, 27), 55: (0, 69), 31: (0, 4), 62: (0, 11)}, 219: {79: (0, 241), 83: (0, 6), 30: (1, {'@': 59})}, 220: {79: (1, {'@': 199}), 78: (1, {'@': 199})}, 221: {0: (1, {'@': 100}), 1: (1, {'@': 100}), 2: (1, {'@': 100}), 3: (1, {'@': 100}), 4: (1, {'@': 100}), 5: (1, {'@': 100}), 6: (1, {'@': 100}), 7: (1, {'@': 100}), 8: (1, {'@': 100}), 9: (1, {'@': 100}), 10: (1, {'@': 100}), 11: (1, {'@': 100}), 12: (1, {'@': 100}), 13: (1, {'@': 100}), 14: (1, {'@': 100}), 15: (1, {'@': 100}), 16: (1, {'@': 100}), 17: (1, {'@': 100}), 18: (1, {'@': 100}), 19: (1, {'@': 100}), 20: (1, {'@': 100}), 21: (1, {'@': 100}), 22: (1, {'@': 100}), 23: (1, {'@': 100}), 24: (1, {'@': 100}), 25: (1, {'@': 100}), 26: (1, {'@': 100}), 27: (1, {'@': 100}), 28: (1, {'@': 100}), 29: (1, {'@': 100})}, 222: {0: (1, {'@': 152}), 19: (1, {'@': 152}), 24: (1, {'@': 144}), 12: (1, {'@': 144}), 26: (1, {'@': 144}), 6: (1, {'@': 144})}, 223: {0: (1, {'@': 164}), 12: (1, {'@': 164}), 4: (1, {'@': 164}), 18: (1, {'@': 164}), 6: (1, {'@': 164}), 19: (1, {'@': 164}), 24: (1, {'@': 164}), 8: (1, {'@': 164}), 26: (1, {'@': 164}), 29: (1, {'@': 164})}, 224: {79: (0, 22), 78: (0, 31)}, 225: {12: (0, 28)}, 226: {0: (1, {'@': 182}), 1: (1, {'@': 182}), 2: (1, {'@': 182}), 3: (1, {'@': 182}), 4: (1, {'@': 182}), 5: (1, {'@': 182}), 6: (1, {'@': 182}), 7: (1, {'@': 182}), 8: (1, {'@': 182}), 9: (1, {'@': 182}), 10: (1, {'@': 182}), 11: (1, {'@': 182}), 12: (1, {'@': 182}), 13: (1, {'@': 182}), 14: (1, {'@': 182}), 15: (1, {'@': 182}), 16: (1, {'@': 182}), 17: (1, {'@': 182}), 18: (1, {'@': 182}), 19: (1, {'@': 182}), 20: (1, {'@': 182}), 21: (1, {'@': 182}), 22: (1, {'@': 182}), 23: (1, {'@': 182}), 24: (1, {'@': 182}), 25: (1, {'@': 182}), 26: (1, {'@': 182}), 27: (1, {'@': 182}), 28: (1, {'@': 182}), 29: (1, {'@': 182})}, 227: {13: (0, 53), 111: (0, 5), 24: (0, 127), 0: (1, {'@': 163}), 19: (1, {'@': 163}), 8: (1, {'@': 163}), 4: (1, {'@': 163}), 18: (1, {'@': 163}), 29: (1, {'@': 163})}, 228: {30: (1, {'@': 60})}, 229: {78: (0, 107), 79: (0, 114)}, 230: {78: (1, {'@': 200}), 79: (1, {'@': 200})}, 231: {0: (1, {'@': 161}), 12: (1, {'@': 161}), 4: (1, {'@': 161}), 18: (1, {'@': 161}), 6: (1, {'@': 161}), 19: (1, {'@': 161}), 24: (1, {'@': 161}), 8: (1, {'@': 161}), 26: (1, {'@': 161})}, 232: {52: (0, 131), 48: (0, 48), 43: (0, 119), 39: (0, 18), 57: (0, 236)}, 233: {12: (0, 126)}, 234: {63: (0, 201), 12: (0, 108), 24: (0, 210)}, 235: {24: (0, 8)}, 236: {0: (1, {'@': 202}), 1: (1, {'@': 202}), 2: (1, {'@': 202}), 3: (1, {'@': 202}), 4: (1, {'@': 202}), 5: (1, {'@': 202}), 6: (1, {'@': 202}), 7: (1, {'@': 202}), 8: (1, {'@': 202}), 9: (1, {'@': 202}), 10: (1, {'@': 202}), 11: (1, {'@': 202}), 12: (1, {'@': 202}), 13: (1, {'@': 202}), 14: (1, {'@': 202}), 15: (1, {'@': 202}), 16: (1, {'@': 202}), 17: (1, {'@': 202}), 18: (1, {'@': 202}), 19: (1, {'@': 202}), 20: (1, {'@': 202}), 21: (1, {'@': 202}), 22: (1, {'@': 202}), 23: (1, {'@': 202}), 24: (1, {'@': 202}), 33: (1, {'@': 202}), 25: (1, {'@': 202}), 27: (1, {'@': 202}), 28: (1, {'@': 202}), 26: (1, {'@': 202}), 29: (1, {'@': 202})}, 237: {30: (1, {'@': 78})}, 238: {30: (1, {'@': 62})}, 239: {54: (1, {'@': 198})}, 240: {0: (1, {'@': 155}), 12: (1, {'@': 155}), 24: (1, {'@': 155}), 8: (1, {'@': 155}), 6: (1, {'@': 155}), 26: (1, {'@': 155}), 19: (1, {'@': 155})}, 241: {95: (0, 220), 115: (0, 229)}, 242: {0: (1, {'@': 210}), 12: (1, {'@': 210}), 6: (1, {'@': 210}), 19: (1, {'@': 210}), 24: (1, {'@': 210}), 8: (1, {'@': 210}), 26: (1, {'@': 210})}, 243: {12: (0, 207)}, 244: {30: (1, {'@': 66})}, 245: {116: (0, 121)}, 246: {54: (0, 40)}, 247: {0: (1, {'@': 158}), 12: (1, {'@': 158}), 4: (1, {'@': 158}), 6: (1, {'@': 158}), 19: (1, {'@': 158}), 24: (1, {'@': 158}), 8: (1, {'@': 158}), 26: (1, {'@': 158})}, 248: {0: (1, {'@': 105}), 1: (1, {'@': 105}), 2: (1, {'@': 105}), 3: (1, {'@': 105}), 4: (1, {'@': 105}), 5: (1, {'@': 105}), 6: (1, {'@': 105}), 7: (1, {'@': 105}), 8: (1, {'@': 105}), 9: (1, {'@': 105}), 10: (1, {'@': 105}), 11: (1, {'@': 105}), 12: (1, {'@': 105}), 13: (1, {'@': 105}), 14: (1, {'@': 105}), 15: (1, {'@': 105}), 16: (1, {'@': 105}), 17: (1, {'@': 105}), 18: (1, {'@': 105}), 19: (1, {'@': 105}), 20: (1, {'@': 105}), 21: (1, {'@': 105}), 22: (1, {'@': 105}), 23: (1, {'@': 105}), 24: (1, {'@': 105}), 25: (1, {'@': 105}), 26: (1, {'@': 105}), 27: (1, {'@': 105}), 28: (1, {'@': 105}), 29: (1, {'@': 105})}, 249: {12: (1, {'@': 178})}, 250: {30: (1, {'@': 184}), 79: (1, {'@': 184})}, 251: {32: (0, 125), 33: (0, 116), 34: (0, 221), 36: (0, 197), 38: (0, 76), 39: (0, 19), 40: (0, 17), 42: (0, 3), 43: (0, 119), 44: (0, 96), 46: (0, 248), 47: (0, 83), 48: (0, 48), 50: (0, 37), 59: (0, 0), 72: (0, 80), 61: (0, 27), 55: (0, 69), 74: (0, 124), 31: (0, 4), 62: (0, 11), 69: (0, 223), 75: (0, 41), 35: (0, 35), 37: (0, 32), 41: (0, 178), 77: (0, 231), 45: (0, 52), 49: (0, 23), 51: (0, 89), 52: (0, 131), 53: (0, 151), 54: (0, 170), 70: (0, 240), 76: (0, 247), 56: (0, 232), 68: (0, 16), 57: (0, 236), 58: (0, 182), 60: (0, 149), 15: (0, 9)}, 252: {12: (0, 137), 27: (0, 13)}, 253: {30: (1, {'@': 74})}, 254: {0: (1, {'@': 132}), 1: (1, {'@': 132}), 2: (1, {'@': 132}), 3: (1, {'@': 132}), 4: (1, {'@': 132}), 5: (1, {'@': 132}), 6: (1, {'@': 132}), 7: (1, {'@': 132}), 8: (1, {'@': 132}), 9: (1, {'@': 132}), 10: (1, {'@': 132}), 11: (1, {'@': 132}), 12: (1, {'@': 132}), 13: (1, {'@': 132}), 14: (1, {'@': 132}), 15: (1, {'@': 132}), 16: (1, {'@': 132}), 17: (1, {'@': 132}), 18: (1, {'@': 132}), 19: (1, {'@': 132}), 20: (1, {'@': 132}), 21: (1, {'@': 132}), 22: (1, {'@': 132}), 23: (1, {'@': 132}), 24: (1, {'@': 132}), 25: (1, {'@': 132}), 26: (1, {'@': 132}), 27: (1, {'@': 132}), 28: (1, {'@': 132}), 29: (1, {'@': 132})}, 255: {0: (1, {'@': 218}), 13: (1, {'@': 218}), 4: (1, {'@': 218}), 18: (1, {'@': 218}), 19: (1, {'@': 218}), 6: (1, {'@': 218}), 24: (1, {'@': 218}), 8: (1, {'@': 218}), 26: (1, {'@': 218}), 29: (1, {'@': 218}), 12: (1, {'@': 218})}, 256: {12: (1, {'@': 171})}, 257: {79: (0, 241), 83: (0, 238), 30: (1, {'@': 63})}, 258: {63: (0, 26), 12: (0, 108), 24: (0, 34)}, 259: {0: (1, {'@': 118}), 1: (1, {'@': 118}), 2: (1, {'@': 118}), 3: (1, {'@': 118}), 4: (1, {'@': 118}), 5: (1, {'@': 118}), 6: (1, {'@': 118}), 7: (1, {'@': 118}), 8: (1, {'@': 118}), 9: (1, {'@': 118}), 10: (1, {'@': 118}), 11: (1, {'@': 118}), 12: (1, {'@': 118}), 13: (1, {'@': 118}), 14: (1, {'@': 118}), 15: (1, {'@': 118}), 16: (1, {'@': 118}), 17: (1, {'@': 118}), 18: (1, {'@': 118}), 19: (1, {'@': 118}), 20: (1, {'@': 118}), 21: (1, {'@': 118}), 22: (1, {'@': 118}), 23: (1, {'@': 118}), 24: (1, {'@': 118}), 25: (1, {'@': 118}), 26: (1, {'@': 118}), 27: (1, {'@': 118}), 28: (1, {'@': 118}), 29: (1, {'@': 118})}, 260: {30: (1, {'@': 70})}, 261: {0: (1, {'@': 133}), 1: (1, {'@': 133}), 2: (1, {'@': 133}), 3: (1, {'@': 133}), 4: (1, {'@': 133}), 5: (1, {'@': 133}), 6: (1, {'@': 133}), 7: (1, {'@': 133}), 8: (1, {'@': 133}), 9: (1, {'@': 133}), 10: (1, {'@': 133}), 11: (1, {'@': 133}), 12: (1, {'@': 133}), 13: (1, {'@': 133}), 14: (1, {'@': 133}), 15: (1, {'@': 133}), 16: (1, {'@': 133}), 17: (1, {'@': 133}), 18: (1, {'@': 133}), 19: (1, {'@': 133}), 20: (1, {'@': 133}), 21: (1, {'@': 133}), 22: (1, {'@': 133}), 23: (1, {'@': 133}), 24: (1, {'@': 133}), 25: (1, {'@': 133}), 26: (1, {'@': 133}), 27: (1, {'@': 133}), 28: (1, {'@': 133}), 29: (1, {'@': 133})}, 262: {0: (1, {'@': 129}), 1: (1, {'@': 129}), 2: (1, {'@': 129}), 3: (1, {'@': 129}), 4: (1, {'@': 129}), 5: (1, {'@': 129}), 6: (1, {'@': 129}), 7: (1, {'@': 129}), 8: (1, {'@': 129}), 9: (1, {'@': 129}), 10: (1, {'@': 129}), 11: (1, {'@': 129}), 12: (1, {'@': 129}), 13: (1, {'@': 129}), 14: (1, {'@': 129}), 15: (1, {'@': 129}), 16: (1, {'@': 129}), 17: (1, {'@': 129}), 18: (1, {'@': 129}), 19: (1, {'@': 129}), 20: (1, {'@': 129}), 21: (1, {'@': 129}), 22: (1, {'@': 129}), 23: (1, {'@': 129}), 24: (1, {'@': 129}), 25: (1, {'@': 129}), 26: (1, {'@': 129}), 27: (1, {'@': 129}), 28: (1, {'@': 129}), 29: (1, {'@': 129})}}, 'start_states': {'start': 49}, 'end_states': {'start': 133}}, 'options': {'debug': False, 'keep_all_tokens': False, 'tree_class': None, 'cache': False, 'postlex': None, 'parser': 'lalr', 'lexer': 'contextual', 'transformer': None, 'start': ['start'], 'priority': 'normal', 'ambiguity': 'auto', 'regex': False, 'propagate_positions': False, 'lexer_callbacks': {}, 'maybe_placeholders': False, 'edit_terminals': None, 'g_regex_flags': 0, 'use_bytes': False, 'import_paths': [], 'source_path': None}, '__type__': 'ParsingFrontend'}, 'rules': [{'@': 55}, {'@': 56}, {'@': 57}, {'@': 58}, {'@': 59}, {'@': 60}, {'@': 61}, {'@': 62}, {'@': 63}, {'@': 64}, {'@': 65}, {'@': 66}, {'@': 67}, {'@': 68}, {'@': 69}, {'@': 70}, {'@': 71}, {'@': 72}, {'@': 73}, {'@': 74}, {'@': 75}, {'@': 76}, {'@': 77}, {'@': 78}, {'@': 79}, {'@': 80}, {'@': 81}, {'@': 82}, {'@': 83}, {'@': 84}, {'@': 85}, {'@': 86}, {'@': 87}, {'@': 88}, {'@': 89}, {'@': 90}, {'@': 91}, {'@': 92}, {'@': 93}, {'@': 94}, {'@': 95}, {'@': 96}, {'@': 97}, {'@': 98}, {'@': 99}, {'@': 100}, {'@': 101}, {'@': 102}, {'@': 103}, {'@': 104}, {'@': 105}, {'@': 106}, {'@': 107}, {'@': 108}, {'@': 109}, {'@': 110}, {'@': 111}, {'@': 112}, {'@': 113}, {'@': 114}, {'@': 115}, {'@': 116}, {'@': 117}, {'@': 118}, {'@': 119}, {'@': 120}, {'@': 121}, {'@': 122}, {'@': 123}, {'@': 124}, {'@': 125}, {'@': 126}, {'@': 127}, {'@': 128}, {'@': 129}, {'@': 130}, {'@': 131}, {'@': 132}, {'@': 133}, {'@': 134}, {'@': 135}, {'@': 136}, {'@': 137}, {'@': 138}, {'@': 139}, {'@': 140}, {'@': 141}, {'@': 142}, {'@': 143}, {'@': 144}, {'@': 145}, {'@': 146}, {'@': 147}, {'@': 148}, {'@': 149}, {'@': 150}, {'@': 151}, {'@': 152}, {'@': 153}, {'@': 154}, {'@': 155}, {'@': 156}, {'@': 157}, {'@': 158}, {'@': 159}, {'@': 160}, {'@': 161}, {'@': 162}, {'@': 163}, {'@': 164}, {'@': 165}, {'@': 166}, {'@': 167}, {'@': 168}, {'@': 169}, {'@': 170}, {'@': 171}, {'@': 172}, {'@': 173}, {'@': 174}, {'@': 175}, {'@': 176}, {'@': 177}, {'@': 178}, {'@': 179}, {'@': 180}, {'@': 181}, {'@': 182}, {'@': 183}, {'@': 184}, {'@': 185}, {'@': 186}, {'@': 187}, {'@': 188}, {'@': 189}, {'@': 190}, {'@': 191}, {'@': 192}, {'@': 193}, {'@': 194}, {'@': 195}, {'@': 196}, {'@': 197}, {'@': 198}, {'@': 199}, {'@': 200}, {'@': 201}, {'@': 202}, {'@': 203}, {'@': 204}, {'@': 205}, {'@': 206}, {'@': 207}, {'@': 208}, {'@': 209}, {'@': 210}, {'@': 211}, {'@': 212}, {'@': 213}, {'@': 214}, {'@': 215}, {'@': 216}, {'@': 217}, {'@': 218}, {'@': 219}], 'options': {'debug': False, 'keep_all_tokens': False, 'tree_class': None, 'cache': False, 'postlex': None, 'parser': 'lalr', 'lexer': 'contextual', 'transformer': None, 'start': ['start'], 'priority': 'normal', 'ambiguity': 'auto', 'regex': False, 'propagate_positions': False, 'lexer_callbacks': {}, 'maybe_placeholders': False, 'edit_terminals': None, 'g_regex_flags': 0, 'use_bytes': False, 'import_paths': [], 'source_path': None}, '__type__': 'Lark'}
)
MEMO = (
{0: {'name': 'NUMBER', 'pattern': {'value': '(?:(?:\\+|\\-))?(?:[0-9])+', 'flags': [], '_width': [1, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 1: {'name': 'WS', 'pattern': {'value': '(?:[ \t\x0c\r\n])+', 'flags': [], '_width': [1, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 2: {'name': 'DOT', 'pattern': {'value': '.', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 3: {'name': 'QMARK', 'pattern': {'value': '?', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 4: {'name': 'BANG', 'pattern': {'value': '!', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 5: {'name': 'AT', 'pattern': {'value': '@', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 6: {'name': 'COMMA', 'pattern': {'value': ',', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 7: {'name': 'LPAR', 'pattern': {'value': '(', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 8: {'name': 'RPAR', 'pattern': {'value': ')', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 9: {'name': 'LESSTHAN', 'pattern': {'value': '<', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 10: {'name': 'MORETHAN', 'pattern': {'value': '>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 11: {'name': '__ANON_0', 'pattern': {'value': '-->', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 12: {'name': '__ANON_1', 'pattern': {'value': '<->', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 13: {'name': '__ANON_2', 'pattern': {'value': '{--', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 14: {'name': '__ANON_3', 'pattern': {'value': '--]', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 15: {'name': '__ANON_4', 'pattern': {'value': '{-]', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 16: {'name': '__ANON_5', 'pattern': {'value': '==>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 17: {'name': '__ANON_6', 'pattern': {'value': '=/>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 18: {'name': '__ANON_7', 'pattern': {'value': '=|>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 19: {'name': '__ANON_8', 'pattern': {'value': '=\\>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 20: {'name': '__ANON_9', 'pattern': {'value': '<=>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 21: {'name': '__ANON_10', 'pattern': {'value': '</>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 22: {'name': '__ANON_11', 'pattern': {'value': '<|>', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 23: {'name': 'CIRCUMFLEX', 'pattern': {'value': '^', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 24: {'name': 'PLUS', 'pattern': {'value': '+', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 25: {'name': 'RSQB', 'pattern': {'value': ']', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 26: {'name': 'RBRACE', 'pattern': {'value': '}', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 27: {'name': 'STAR', 'pattern': {'value': '*', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 28: {'name': '__ANON_12', 'pattern': {'value': '||', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 29: {'name': '__ANON_13', 'pattern': {'value': '&&', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 30: {'name': '__ANON_14', 'pattern': {'value': '&/', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 31: {'name': '__ANON_15', 'pattern': {'value': '&|', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 32: {'name': 'VBAR', 'pattern': {'value': '|', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 33: {'name': 'AMPERSAND', 'pattern': {'value': '&', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 34: {'name': 'MINUS', 'pattern': {'value': '-', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 35: {'name': 'TILDE', 'pattern': {'value': '~', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 36: {'name': 'LSQB', 'pattern': {'value': '[', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 37: {'name': 'LBRACE', 'pattern': {'value': '{', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 38: {'name': '__ANON_16', 'pattern': {'value': '--', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 39: {'name': 'BACKSLASH', 'pattern': {'value': '\\', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 40: {'name': 'SLASH', 'pattern': {'value': '/', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 41: {'name': 'DOLLAR', 'pattern': {'value': '$', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 42: {'name': 'HASH', 'pattern': {'value': '#', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 43: {'name': '__ANON_17', 'pattern': {'value': ':!', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 44: {'name': 'COLON', 'pattern': {'value': ':', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 45: {'name': '__ANON_18', 'pattern': {'value': ':/:', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 46: {'name': '__ANON_19', 'pattern': {'value': ':|:', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 47: {'name': '__ANON_20', 'pattern': {'value': ':\\:', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 48: {'name': 'SEMICOLON', 'pattern': {'value': ';', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 49: {'name': 'PERCENT', 'pattern': {'value': '%', 'flags': [], '__type__': 'PatternStr'}, 'priority': 1, '__type__': 'TerminalDef'}, 50: {'name': '__ANON_21', 'pattern': {'value': '([0]?\\.[0-9]+|1\\.[0]*|1|0)', 'flags': [], '_width': [1, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 51: {'name': '__ANON_22', 'pattern': {'value': '[0]?\\.[0]*[1-9]{1}[0-9]*', 'flags': [], '_width': [2, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 52: {'name': '__ANON_23', 'pattern': {'value': '[1-9]{1}[0-9]*', 'flags': [], '_width': [1, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 53: {'name': '__ANON_24', 'pattern': {'value': '"[^"]+"', 'flags': [], '_width': [3, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 54: {'name': '__ANON_25', 'pattern': {'value': '[^\\-^\\+^<^>^=^"^&^|^!^.^?^@^~^%^;^\\,^:^\\/^\\\\^*^#^$^\\[^\\]^\\{^\\}^\\(^\\)^\\ ]+', 'flags': [], '_width': [1, 4294967295], '__type__': 'PatternRE'}, 'priority': 1, '__type__': 'TerminalDef'}, 55: {'origin': {'name': 'start', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'task', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 56: {'origin': {'name': 'task', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'budget', '__type__': 'NonTerminal'}, {'name': 'sentence', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 57: {'origin': {'name': 'task', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'sentence', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': [True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 58: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}, {'name': 'truth', '__type__': 'NonTerminal'}], 'order': 0, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 59: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 1, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 60: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'truth', '__type__': 'NonTerminal'}], 'order': 2, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 61: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 62: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}, {'name': 'truth', '__type__': 'NonTerminal'}], 'order': 4, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 63: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 5, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 64: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'truth', '__type__': 'NonTerminal'}], 'order': 6, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 65: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'DOT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 7, 'alias': 'judgement', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 66: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'QMARK', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 8, 'alias': 'question', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 67: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'QMARK', 'filter_out': True, '__type__': 'Terminal'}], 'order': 9, 'alias': 'question', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 68: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'QMARK', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 10, 'alias': 'question', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 69: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'QMARK', 'filter_out': True, '__type__': 'Terminal'}], 'order': 11, 'alias': 'question', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 70: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}, {'name': 'desire', '__type__': 'NonTerminal'}], 'order': 12, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 71: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 13, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 72: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'desire', '__type__': 'NonTerminal'}], 'order': 14, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 73: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}], 'order': 15, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 74: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}, {'name': 'desire', '__type__': 'NonTerminal'}], 'order': 16, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 75: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 17, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 76: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'desire', '__type__': 'NonTerminal'}], 'order': 18, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 77: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'BANG', 'filter_out': True, '__type__': 'Terminal'}], 'order': 19, 'alias': 'goal', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 78: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'AT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 20, 'alias': 'quest', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 79: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}, {'name': 'AT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 21, 'alias': 'quest', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 80: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'AT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'tense', '__type__': 'NonTerminal'}], 'order': 22, 'alias': 'quest', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 81: {'origin': {'name': 'sentence', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}, {'name': 'AT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 23, 'alias': 'quest', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': [False, False, True], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 82: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LESSTHAN', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'copula', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'MORETHAN', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 83: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'copula', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 84: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'op', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'statement_operation1', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 85: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'op', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'statement_operation1', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 86: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'word', '__type__': 'NonTerminal'}, {'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 4, 'alias': 'statement_operation2', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 87: {'origin': {'name': 'statement', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'word', '__type__': 'NonTerminal'}, {'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 5, 'alias': 'statement_operation2', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 88: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_0', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'inheritance', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 89: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_1', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'similarity', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 90: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_2', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'instance', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 91: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_3', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'property', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 92: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_4', 'filter_out': True, '__type__': 'Terminal'}], 'order': 4, 'alias': 'instance_property', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 93: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_5', 'filter_out': True, '__type__': 'Terminal'}], 'order': 5, 'alias': 'implication', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 94: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_6', 'filter_out': True, '__type__': 'Terminal'}], 'order': 6, 'alias': 'predictive_implication', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 95: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_7', 'filter_out': True, '__type__': 'Terminal'}], 'order': 7, 'alias': 'concurrent_implication', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 96: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_8', 'filter_out': True, '__type__': 'Terminal'}], 'order': 8, 'alias': 'retrospective_implication', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 97: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_9', 'filter_out': True, '__type__': 'Terminal'}], 'order': 9, 'alias': 'equivalence', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 98: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_10', 'filter_out': True, '__type__': 'Terminal'}], 'order': 10, 'alias': 'predictive_equivalence', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 99: {'origin': {'name': 'copula', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_11', 'filter_out': True, '__type__': 'Terminal'}], 'order': 11, 'alias': 'concurrent_equivalence', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 100: {'origin': {'name': 'term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'variable', '__type__': 'NonTerminal'}], 'order': 0, 'alias': 'variable_term', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 101: {'origin': {'name': 'term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term_nonvar', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 102: {'origin': {'name': 'term_nonvar', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'interval', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 103: {'origin': {'name': 'term_nonvar', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'word', '__type__': 'NonTerminal'}], 'order': 1, 'alias': 'atom_term', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 104: {'origin': {'name': 'term_nonvar', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'compound_term', '__type__': 'NonTerminal'}], 'order': 2, 'alias': 'compound_term', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 105: {'origin': {'name': 'term_nonvar', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'statement', '__type__': 'NonTerminal'}], 'order': 3, 'alias': 'statement_term', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 106: {'origin': {'name': 'term_nonvar', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'op', '__type__': 'NonTerminal'}], 'order': 4, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 107: {'origin': {'name': 'op', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'CIRCUMFLEX', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'word', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 108: {'origin': {'name': 'interval', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'PLUS', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'NUMBER', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 109: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'set', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 110: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 111: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'single', '__type__': 'NonTerminal'}], 'order': 2, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 112: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'ext_image', '__type__': 'NonTerminal'}], 'order': 3, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 113: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'int_image', '__type__': 'NonTerminal'}], 'order': 4, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 114: {'origin': {'name': 'compound_term', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'negation', '__type__': 'NonTerminal'}], 'order': 5, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 115: {'origin': {'name': 'set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'int_set', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 116: {'origin': {'name': 'set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'ext_set', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 117: {'origin': {'name': 'int_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'con_int_set', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RSQB', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'set', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 118: {'origin': {'name': 'int_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'con_int_set', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RSQB', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'set', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 119: {'origin': {'name': 'ext_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'con_ext_set', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RBRACE', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'set', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 120: {'origin': {'name': 'ext_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'con_ext_set', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RBRACE', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'set', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 121: {'origin': {'name': 'negation', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'con_negation', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 122: {'origin': {'name': 'negation', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_negation', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 123: {'origin': {'name': 'int_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_int_image', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 124: {'origin': {'name': 'int_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_int_image', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 125: {'origin': {'name': 'ext_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_ext_image', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 126: {'origin': {'name': 'ext_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_ext_image', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 127: {'origin': {'name': 'multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_multi', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'multi_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 128: {'origin': {'name': 'multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 129: {'origin': {'name': 'multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'multi_prefix_product', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 130: {'origin': {'name': 'multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_product', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'multi_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 131: {'origin': {'name': 'multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_product', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 4, 'alias': 'multi_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 132: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'single_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 133: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'single_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 134: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'single_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 135: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'single_prefix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 136: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 4, 'alias': 'single_infix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 137: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 5, 'alias': 'single_infix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 138: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'term', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 6, 'alias': 'single_infix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 139: {'origin': {'name': 'single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LPAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'con_single', '__type__': 'NonTerminal'}, {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, {'name': 'RPAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 7, 'alias': 'single_infix', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 140: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_extint_expr', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 141: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_intint_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 142: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_parallel_expr', '__type__': 'NonTerminal'}], 'order': 2, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 143: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_sequential_expr', '__type__': 'NonTerminal'}], 'order': 3, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 144: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_conj_expr', '__type__': 'NonTerminal'}], 'order': 4, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 145: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_disj_expr', '__type__': 'NonTerminal'}], 'order': 5, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 146: {'origin': {'name': 'multi_infix_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_prod_expr', '__type__': 'NonTerminal'}], 'order': 6, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 147: {'origin': {'name': 'multi_prod_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term6', '__type__': 'NonTerminal'}, {'name': '__multi_prod_expr_plus_1', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 148: {'origin': {'name': 'term6', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term5', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 149: {'origin': {'name': 'term6', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_disj_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 150: {'origin': {'name': 'multi_disj_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term5', '__type__': 'NonTerminal'}, {'name': '__multi_disj_expr_plus_2', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 151: {'origin': {'name': 'term5', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term4', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 152: {'origin': {'name': 'term5', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_conj_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 153: {'origin': {'name': 'multi_conj_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term4', '__type__': 'NonTerminal'}, {'name': '__multi_conj_expr_plus_3', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 154: {'origin': {'name': 'term4', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term3', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 155: {'origin': {'name': 'term4', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_sequential_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 156: {'origin': {'name': 'multi_sequential_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term3', '__type__': 'NonTerminal'}, {'name': '__multi_sequential_expr_plus_4', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 157: {'origin': {'name': 'term3', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term2', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 158: {'origin': {'name': 'term3', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_parallel_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 159: {'origin': {'name': 'multi_parallel_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term2', '__type__': 'NonTerminal'}, {'name': '__multi_parallel_expr_plus_5', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 160: {'origin': {'name': 'term2', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term1', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 161: {'origin': {'name': 'term2', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_intint_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 162: {'origin': {'name': 'multi_intint_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term1', '__type__': 'NonTerminal'}, {'name': '__multi_intint_expr_plus_6', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 163: {'origin': {'name': 'term1', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 164: {'origin': {'name': 'term1', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'multi_extint_expr', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 165: {'origin': {'name': 'multi_extint_expr', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'term', '__type__': 'NonTerminal'}, {'name': '__multi_extint_expr_plus_7', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 166: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_13', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'con_conjunction', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 167: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_12', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'con_disjunction', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 168: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_15', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'con_parallel_events', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 169: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_14', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'con_sequential_events', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 170: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'VBAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 4, 'alias': 'con_intensional_intersection', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 171: {'origin': {'name': 'con_multi', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'AMPERSAND', 'filter_out': True, '__type__': 'Terminal'}], 'order': 5, 'alias': 'con_extensional_intersection', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 172: {'origin': {'name': 'con_product', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'STAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 173: {'origin': {'name': 'con_single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'MINUS', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'con_extensional_difference', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 174: {'origin': {'name': 'con_single', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'TILDE', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'con_intensional_difference', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 175: {'origin': {'name': 'con_int_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LSQB', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 176: {'origin': {'name': 'con_ext_set', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'LBRACE', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 177: {'origin': {'name': 'con_negation', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_16', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 178: {'origin': {'name': 'con_int_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'BACKSLASH', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 179: {'origin': {'name': 'con_ext_image', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'SLASH', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 180: {'origin': {'name': 'variable', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'word', '__type__': 'NonTerminal'}], 'order': 0, 'alias': 'independent_var', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 181: {'origin': {'name': 'variable', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'HASH', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'word', '__type__': 'NonTerminal'}], 'order': 1, 'alias': 'dependent_var', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 182: {'origin': {'name': 'variable', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'QMARK', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'word', '__type__': 'NonTerminal'}], 'order': 2, 'alias': 'query_var', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': 0, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 183: {'origin': {'name': 'tense', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_17', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'NUMBER', 'filter_out': False, '__type__': 'Terminal'}, {'name': 'COLON', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': 'tense_time', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 184: {'origin': {'name': 'tense', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_18', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': 'tense_future', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 185: {'origin': {'name': 'tense', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_19', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': 'tense_present', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 186: {'origin': {'name': 'tense', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_20', 'filter_out': True, '__type__': 'Terminal'}], 'order': 3, 'alias': 'tense_past', 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 187: {'origin': {'name': 'desire', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'truth', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 188: {'origin': {'name': 'truth', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'frequency', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'confidence', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'k_evidence', '__type__': 'NonTerminal'}, {'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 189: {'origin': {'name': 'truth', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'frequency', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'confidence', '__type__': 'NonTerminal'}, {'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': [False, False, False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 190: {'origin': {'name': 'truth', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'frequency', '__type__': 'NonTerminal'}, {'name': 'PERCENT', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 191: {'origin': {'name': 'budget', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'priority', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'durability', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'quality', '__type__': 'NonTerminal'}, {'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': 2, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 192: {'origin': {'name': 'budget', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'priority', '__type__': 'NonTerminal'}, {'name': 'SEMICOLON', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'durability', '__type__': 'NonTerminal'}, {'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': 2, 'template_source': None, 'empty_indices': [False, False, False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 193: {'origin': {'name': 'budget', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'priority', '__type__': 'NonTerminal'}, {'name': 'DOLLAR', 'filter_out': True, '__type__': 'Terminal'}], 'order': 2, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': 2, 'template_source': None, 'empty_indices': [False, False, True, False], '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 194: {'origin': {'name': 'word', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'string_raw', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 195: {'origin': {'name': 'word', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'string', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 196: {'origin': {'name': 'priority', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_21', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 197: {'origin': {'name': 'durability', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_22', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 198: {'origin': {'name': 'quality', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_21', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 199: {'origin': {'name': 'frequency', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_21', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 200: {'origin': {'name': 'confidence', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_22', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 201: {'origin': {'name': 'k_evidence', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_23', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 202: {'origin': {'name': 'string', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_24', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 203: {'origin': {'name': 'string_raw', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_25', 'filter_out': False, '__type__': 'Terminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': True, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 204: {'origin': {'name': '__statement_star_0', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 205: {'origin': {'name': '__statement_star_0', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__statement_star_0', '__type__': 'NonTerminal'}, {'name': 'COMMA', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 206: {'origin': {'name': '__multi_prod_expr_plus_1', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'STAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term6', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 207: {'origin': {'name': '__multi_prod_expr_plus_1', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_prod_expr_plus_1', '__type__': 'NonTerminal'}, {'name': 'STAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term6', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 208: {'origin': {'name': '__multi_disj_expr_plus_2', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_12', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term5', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 209: {'origin': {'name': '__multi_disj_expr_plus_2', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_disj_expr_plus_2', '__type__': 'NonTerminal'}, {'name': '__ANON_12', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term5', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 210: {'origin': {'name': '__multi_conj_expr_plus_3', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_13', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term4', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 211: {'origin': {'name': '__multi_conj_expr_plus_3', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_conj_expr_plus_3', '__type__': 'NonTerminal'}, {'name': '__ANON_13', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term4', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 212: {'origin': {'name': '__multi_sequential_expr_plus_4', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_14', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term3', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 213: {'origin': {'name': '__multi_sequential_expr_plus_4', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_sequential_expr_plus_4', '__type__': 'NonTerminal'}, {'name': '__ANON_14', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term3', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 214: {'origin': {'name': '__multi_parallel_expr_plus_5', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__ANON_15', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term2', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 215: {'origin': {'name': '__multi_parallel_expr_plus_5', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_parallel_expr_plus_5', '__type__': 'NonTerminal'}, {'name': '__ANON_15', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term2', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 216: {'origin': {'name': '__multi_intint_expr_plus_6', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'VBAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term1', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 217: {'origin': {'name': '__multi_intint_expr_plus_6', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_intint_expr_plus_6', '__type__': 'NonTerminal'}, {'name': 'VBAR', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term1', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 218: {'origin': {'name': '__multi_extint_expr_plus_7', '__type__': 'NonTerminal'}, 'expansion': [{'name': 'AMPERSAND', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}], 'order': 0, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}, 219: {'origin': {'name': '__multi_extint_expr_plus_7', '__type__': 'NonTerminal'}, 'expansion': [{'name': '__multi_extint_expr_plus_7', '__type__': 'NonTerminal'}, {'name': 'AMPERSAND', 'filter_out': True, '__type__': 'Terminal'}, {'name': 'term', '__type__': 'NonTerminal'}], 'order': 1, 'alias': None, 'options': {'keep_all_tokens': False, 'expand1': False, 'priority': None, 'template_source': None, 'empty_indices': (), '__type__': 'RuleOptions'}, '__type__': 'Rule'}}
)
Shift = 0
Reduce = 1
def Lark_StandAlone(**kwargs):
  return Lark._load_from_dict(DATA, MEMO, **kwargs)
